
from astropy.io import ascii
from astropy.table import MaskedColumn, Table
import logging
import math
import numpy
import os

import daophot
from downloads.cutouts.downloader import ImageDownloader
import util

BRIGHT_LIMIT = 23.0
OBJECT_PLANTED = "Object.planted"
MINIMUM_BRIGHT_DETECTIONS = 5
MINIMUM_BRIGHT_FRACTION = 0.5


def measure_mags(measures):
    """
    Given a list of readings compute the magnitudes for all sources in each reading.

    @param measures: list of readings
    @return: None
    """

    image_downloader = ImageDownloader()

    observations = {}
    for measure in measures:
        for reading in measure:
            if reading.obs not in observations:
                observations[reading.obs] = {'x': [],
                                             'y': [],
                                             'source': image_downloader.download(reading, needs_apcor=True)}
            observations[reading.obs]['x'].append(reading.x)
            observations[reading.obs]['y'].append(reading.y)

    for observation in observations:
        source = observations[observation]['source']
        observations[observation]['mags'] = daophot.phot(source._hdulist_on_disk(),
                                                         observations[observation]['x'],
                                                         observations[observation]['y'],
                                                         aperture=source.apcor.aperture,
                                                         sky=source.apcor.sky,
                                                         swidth=source.apcor.swidth,
                                                         apcor=source.apcor.apcor,
                                                         zmag=source.zmag,
                                                         maxcount=30000,
                                                         extno=source._ext)

    return observations


def match_planted(fk_candidate_observations, match_filename, bright_limit=BRIGHT_LIMIT, object_planted=OBJECT_PLANTED,
                  minimum_bright_detections=MINIMUM_BRIGHT_DETECTIONS, bright_fraction=MINIMUM_BRIGHT_FRACTION):
    """
    Using the fk_candidate_observations as input get the Object.planted file from VOSpace and match
    planted sources with found sources.

    The Object.planted list is pulled from VOSpace based on the standard file-layout and name of the
    first exposure as read from the .astrom file.

    :param fk_candidate_observations: name of the fk*reals.astrom file to check against Object.planted
    :param match_filename: a file that will contain a list of all planted sources and the matched found source

    """

    found_pos = []
    detections = fk_candidate_observations.get_sources()
    for detection in detections:
        reading = detection.get_reading(0)
        # create a list of positions, to be used later by match_lists
        found_pos.append([reading.x, reading.y])

    # Now get the Object.planted file, either from the local FS or from VOSpace.
    objects_planted_uri = object_planted
    if not os.access(objects_planted_uri, os.F_OK):
        objects_planted_uri = fk_candidate_observations.observations[0].get_object_planted_uri()
    lines = open(objects_planted_uri).read()

    # we are changing the format of the Object.planted header to be compatible with astropy.io.ascii but
    # there are some old Object.planted files out there so we do these string/replace calls to reset those.
    new_lines = lines.replace("pix rate", "pix_rate")
    new_lines = new_lines.replace("""''/h rate""", "sky_rate")
    planted_objects_table = ascii.read(new_lines, header_start=-1, data_start=0)
    planted_objects_table.meta = None
    # The match_list method expects a list that contains a position, not an x and a y vector, so we transpose.
    planted_pos = numpy.transpose([planted_objects_table['x'].data, planted_objects_table['y'].data])
    # match_idx is an order list.  The list is in the order of the first list of positions and each entry
    # is the index of the matching position from the second list.
    (match_idx, match_fnd) = util.match_lists(numpy.array(planted_pos), numpy.array(found_pos))
    assert isinstance(match_idx, numpy.ma.MaskedArray)
    assert isinstance(match_fnd, numpy.ma.MaskedArray)
    false_positives_table = Table()
    # Once we've matched the two lists we'll need some new columns to store the information in.
    # these are masked columns so that object.planted entries that have no detected match are left 'blank'.
    new_columns = [MaskedColumn(name="measure_x", length=len(planted_objects_table), mask=True),
                   MaskedColumn(name="measure_y", length=len(planted_objects_table), mask=True),
                   MaskedColumn(name="measure_rate", length=len(planted_objects_table), mask=True),
                   MaskedColumn(name="measure_angle", length=len(planted_objects_table), mask=True),
                   MaskedColumn(name="measure_mag1", length=len(planted_objects_table), mask=True),
                   MaskedColumn(name="measure_merr1", length=len(planted_objects_table), mask=True),
                   MaskedColumn(name="measure_mag2", length=len(planted_objects_table), mask=True),
                   MaskedColumn(name="measure_merr2", length=len(planted_objects_table), mask=True),
                   MaskedColumn(name="measure_mag3", length=len(planted_objects_table), mask=True),
                   MaskedColumn(name="measure_merr3", length=len(planted_objects_table), mask=True)]
    planted_objects_table.add_columns(new_columns)
    tlength = 0
    new_columns = [MaskedColumn(name="measure_x", length=tlength, mask=True),
                   MaskedColumn(name="measure_y", length=tlength, mask=True),
                   MaskedColumn(name="measure_rate", length=0, mask=True),
                   MaskedColumn(name="measure_angle", length=0, mask=True),
                   MaskedColumn(name="measure_mag1", length=0, mask=True),
                   MaskedColumn(name="measure_merr1", length=0, mask=True),
                   MaskedColumn(name="measure_mag2", length=0, mask=True),
                   MaskedColumn(name="measure_merr2", length=0, mask=True),
                   MaskedColumn(name="measure_mag3", length=tlength, mask=True),
                   MaskedColumn(name="measure_merr3", length=tlength, mask=True)]
    false_positives_table.add_columns(new_columns)

    # We do some 'checks' on the Object.planted match to diagnose pipeline issues.  Those checks are made using just
    # those planted sources we should have detected.
    bright = planted_objects_table['mag'] < bright_limit
    n_bright_planted = numpy.count_nonzero(planted_objects_table['mag'][bright])

    measures = []
    idxs  = []
    for idx in range(len(match_idx)):
        # The match_idx value is False if nothing was found.
        if not match_idx.mask[idx]:
            # Each 'source' has multiple 'readings'
            measures.append(detections[match_idx[idx]].get_readings())
            idxs.append(idx)

    observations = measure_mags(measures)

    for oidx in range(len(measures)):
        idx = idxs[oidx]
        readings = measures[oidx]
        start_jd = util.Time(readings[0].obs.header['MJD_OBS_CENTER'], format='mpc', scale='utc').jd
        end_jd = util.Time(readings[-1].obs.header['MJD_OBS_CENTER'], format='mpc', scale='utc').jd
        rate = math.sqrt((readings[-1].x - readings[0].x) ** 2 + (readings[-1].y - readings[0].y) ** 2) / (
            24 * (end_jd - start_jd))
        rate = int(rate * 100) / 100.0
        angle = math.degrees(math.atan2(readings[-1].y - readings[0].y, readings[-1].x - readings[0].x))
        angle = int(angle * 100) / 100.0
        planted_objects_table[idx]['measure_rate'] = rate
        planted_objects_table[idx]['measure_angle'] = angle
        planted_objects_table[idx]['measure_x'] = observations[readings[0].obs]['mags']["XCENTER"][oidx]
        planted_objects_table[idx]['measure_y'] = observations[readings[0].obs]['mags']["YCENTER"][oidx]
        for ridx in range(len(readings)):
            reading = readings[ridx]
            mags = observations[reading.obs]['mags']
            planted_objects_table[idx]['measure_mag{}'.format(ridx+1)] = mags["MAG"][oidx]
            planted_objects_table[idx]['measure_merr{}'.format(ridx+1)] = mags["MERR"][oidx]

    # for idx in range(len(match_fnd)):
    #     if match_fnd.mask[idx]:
    #         measures = detections[idx].get_readings()
    #         false_positives_table.add_row()
    #         false_positives_table[-1] = measure_mags(measures, false_positives_table[-1])

    # Count an object as detected if it has a measured magnitude in the first frame of the triplet.
    n_bright_found = numpy.count_nonzero(planted_objects_table['measure_mag1'][bright])
    # Also compute the offset and standard deviation of the measured magnitude from that planted ones.
    offset = numpy.mean(planted_objects_table['mag'][bright] - planted_objects_table['measure_mag1'][bright])
    try:
        offset = "{:5.2f}".format(offset)
    except:
        offset = "indef"

    std = numpy.std(planted_objects_table['mag'][bright] - planted_objects_table['measure_mag1'][bright])
    try:
        std = "{:5.2f}".format(std)
    except:
        std = "indef"

    if os.access(match_filename, os.R_OK):
        fout = open(match_filename, 'a')
    else:
        fout = open(match_filename, 'w')

    fout.write("#K {:10s} {:10s}\n".format("EXPNUM", "FWHM"))
    for measure in detections[0].get_readings():
        fout.write('#V {:10s} {:10s}\n'.format(measure.obs.header['EXPNUM'], measure.obs.header['FWHM']))

    fout.write("#K ")
    for keyword in ["RMIN", "RMAX", "ANGLE", "AWIDTH"]:
        fout.write("{:10s} ".format(keyword))
    fout.write("\n")

    fout.write("#V ")
    for keyword in ["RMIN", "RMAX", "ANGLE", "AWIDTH"]:
        fout.write("{:10s} ".format(fk_candidate_observations.sys_header[keyword]))
    fout.write("\n")

    fout.write("#K ")
    for keyword in ["NBRIGHT", "NFOUND", "OFFSET", "STDEV"]:
        fout.write("{:10s} ".format(keyword))
    fout.write("\n")
    fout.write("#V {:<10} {:<10} {:<10} {:<10}\n".format(n_bright_planted,
                                                         n_bright_found,
                                                         offset,
                                                         std))

    try:
        writer = ascii.FixedWidth
        # add a hash to the start of line that will have header columns: for JMP
        fout.write("#")
        fout.flush()
        ascii.write(planted_objects_table, output=fout, Writer=writer, delimiter=None)
        if len(false_positives_table) > 0:
            with open(match_filename+".fp", 'a') as fpout:
                fpout.write("#")
                ascii.write(false_positives_table, output=fpout, Writer=writer, delimiter=None)
    except Exception as e:
        logging.error(str(e))
        raise e
    finally:
        fout.close()

    # Some simple checks to report a failure how we're doing.
    if n_bright_planted < minimum_bright_detections:
        raise RuntimeError(1, "Too few bright objects planted.")

    if n_bright_found / float(n_bright_planted) < bright_fraction:
        raise RuntimeError(2, "Too few bright objects found.")

    return "{} {} {} {}".format(n_bright_planted, n_bright_found, offset, std)




__author__ = 'Michele Bannister   git:@mtbannister'

import sys

import matplotlib.pyplot as plt
import numpy

from ossos.mpc import TNOdbComment
import parsers
# Define some CSS to control our custom labels
css = """
table
{
  border-collapse: collapse;
}
th
{
  color: #ffffff;
  background-color: #000000;
}
td
{
  background-color: #cccccc;
}
table, th, td
{
  font-family:Sofia Pro, sans-serif;
  border: 1px solid black;
  text-align: right;
}
"""


def all_clean_phot(obj):
    try:
        tno = parsers.ossos_discoveries(single_object=obj, no_nt_and_u=False)[0]
    except:
        return [], [], []
    # print len(tno.orbit.observations)

    # consider only the cleanest photometry: nothing with involvement or other weirdness
    tnoobs = []
    for n in tno.orbit.observations:
        if isinstance(n.comment, TNOdbComment):
            if n.comment.photometry_note == 'Y' and (n.comment.mpc_note == ' ' or n.comment.mpc_note == ''):
                tnoobs.append(n)
    if len(tnoobs) > 0:  # there better be some observations that are untainted
        tno_mags = [n.comment.mag for n in tnoobs]
        mag_dates = [n.date.jd for n in tnoobs]
        mag_err = [n.comment.mag_uncertainty for n in tnoobs]
        return tno_mags, mag_err, mag_dates
    else:
        print 'no unflagged photometry exists for {}'.format(obj)
        return [], [], []


def stddev_phot(obj):
    tno_mags, mag_err, mag_dates = all_clean_phot(obj)
    # without additional parameters, this returns the population std deviation.
    retval = numpy.std(numpy.array(tno_mags))
    # print obj, retval

    return retval


def plot_all_obs_single_object(obj):
    tno_mags, mag_err, mag_dates = all_clean_phot(obj)

    plt.figure()
    plt.errorbar(mag_dates, tno_mags, yerr=mag_err, fmt='.', )
    plt.gca().invert_yaxis()
    plt.title(obj)
    plt.draw()
    plt.show()


if __name__ == '__main__':
    obj = sys.argv[1]
    plot_all_obs_single_object(obj)

    fig, ax = plt.subplots(1, 1, figsize=(35, 10))

    variation = []
    objects = parsers.ossos_discoveries()
    labels = []
    for tno in objects:  # [n for n in objects if n.name in parameters.COLOSSOS]:
        tno_mags, mag_err, mag_dates = all_clean_phot(obj)
        if len(tno_mags) > 0:
            tnomax = max(tno_mags)
            tnomin = min(tno_mags)
            variation.append(tno_mags)
            html = '<table><tr><th>Name</th><td>{:}</td></tr></table>'.format(tno.name)
            # '<tr><th>Dist</th><td>{:.2f}</td></tr>' \
            # '<tr><th>Peri</th><td>{:.2f}</td></tr>
            labels.append(tno.name)
        else:
            print("{} has no clean observations.\n".format(tno.name))

    # points = ax.errorbar(range(0, len(labels)), variation, fmt='.', ecolor='b')
    points = ax.boxplot(variation, labels=labels, showbox=False, showcaps=False, whiskerprops={'linestyle': '-'})
    ax.grid(True, alpha=0.3)
    ax.set_ylabel('$m_{r}$')
    plt.gca().invert_yaxis()
    plt.setp(ax.get_xticklabels(), rotation=90)
    plt.draw()
    outfile = 'All-OSSOS-variation.pdf'
    plt.savefig(outfile, transparent=True, bbox_inches='tight')

    # # Activate the interactivity
    # tooltip = plugins.PointHTMLTooltip(points[0], labels, voffset=10, hoffset=10, css=css)
    # plugins.connect(fig, tooltip)
    # mpld3.show()


    # FIXME: want to add a plot
    #  play with standard deviation of mag of objects vs H vs r_helio



__author__ = 'jjk'

import numpy as np
import os

path = '/Users/bannisterm/Dropbox/Papers in progress/OSSOS/First_quarter/data/'


def square_fit(m, params):
    return np.where(m < 21., params[0],
                    (params[0] - params[1] * (m - 21.0) ** 2) / (1. + np.exp((m - params[2]) / params[3])))


def square_fit_discovery_mag(obj, discov_mag, discov_rate):
    pwd = path + 'efficiency_motion_rates'
    blocks = {'o3e': "13AE", 'o3o': "13AO"}
    block = blocks[obj.lstrip('u')[0:3]]
    smooth_parameter_files = filter(lambda name: name.startswith('smooth'), os.listdir('{}/{}/'.format(pwd, block)))
    smooth_parameter_files.sort(key=lambda x: float(x.split('-')[1]))
    rates = np.array([float(x.split('-')[2]) for x in smooth_parameter_files])
    rate, = np.where(discov_rate < rates)
    fn = smooth_parameter_files[rate[0]]
    pd, ps, mag_limit = read_smooth_fit('{}/{}/{}'.format(pwd, block, fn))
    eta = square_fit(discov_mag, ps)
    # print obj, discov_mag, eta, discov_rate, block, fn

    return float(eta)


def read_smooth_fit(fichier):
    pd, ps, mag_limit = None, None, None
    with open(fichier, 'r') as infile:
        lines = infile.readlines()
        for line in lines:
            if line[0] != "#":
                if line.startswith("double_param="):
                    pd = map(float, line[13:-1].split())
                if line.startswith("square_param="):
                    ps = map(float, line[13:-1].split())
                if line.startswith("mag_lim="):
                    mag_limit = float(line.split()[1])

    return pd, ps, mag_limit



__author__ = "David Rusk <drusk@uvic.ca>"

from matplotlib.colors import LinearSegmentedColormap


class GrayscaleColorMap(object):
    """
    An image color map, which allows its contrast and bias to be controlled.

    Refer to http://matplotlib.org/api/colors_api.html,
    specifically class matplotlib.colors.LinearSegmentedColormap
    to better understand how this works.

    For an example, also see
    http://matplotlib.org/examples/pylab_examples/custom_cmap.html?highlight=codex%20colormap

    For explanation of contrast and bias as in ds9, see:
    http://chandra-ed.harvard.edu/learning_ds9_page2.html
    """

    def __init__(self):
        self.set_defaults()

    def _build_cdict(self):
        lower_segment_x = self._clip(self.x_offset + 0.5 - self.x_spread / 2)
        upper_segment_x = self._clip(self.x_offset + 0.5 + self.x_spread / 2)

        max_y = self._clip(0.5 + (self.y_spread / 2))
        min_y = self._clip(0.5 - (self.y_spread / 2))

        if lower_segment_x > upper_segment_x:
            # If they try to go past maximum contrast, flip the colorbar
            # ds9 style
            lower_segment_x, upper_segment_x = upper_segment_x, lower_segment_x
            min_y, max_y = max_y, min_y

        # NOTE: max_y used at lower bounds and min_y used at upper bounds in
        # order to invert the image's colourmap by default.
        self.min_bounds = (0.0, max_y, max_y)
        self.lower_segment_bounds = (lower_segment_x, max_y, max_y)
        self.upper_segment_bounds = (upper_segment_x, min_y, min_y)
        self.max_bounds = (1.0, min_y, min_y)

        self.cdict = {}
        for color in ["red", "green", "blue"]:
            self.cdict[color] = [self.min_bounds, self.lower_segment_bounds,
                                 self.upper_segment_bounds, self.max_bounds]

    def set_defaults(self):
        self._contrast = 0.5
        self._bias = 0.5

        self.x_spread = 1.0
        self.y_spread = 1.0
        self.x_offset = 0.0

        self._build_cdict()

    def set_bias(self, bias):
        """
        Adjusts the image bias.

        Bias determines where the color changes start.  At low bias, low
        intensities (i.e., low pixel values) will have non-zero color
        differences, while at high bias only high pixel values will have
        non-zero differences

        Args:
          bias: float
            A number between 0 and 1.  Note that upon initialization the
            colormap has a default bias of 0.5.

        Returns: void
        """
        self.x_offset += (bias - self._bias)
        self._bias = bias

        self._build_cdict()

    def update_bias(self, bias_diff):
        self.set_bias(self._bias + bias_diff)

    def set_contrast(self, contrast):
        """
        Adjusts the image contrast.

        Contrast refers to the rate of change of color with color level.
        At low contrast, color changes gradually over many intensity
        levels, while at high contrast it can change rapidly within a
        few levels

        Args:
          contrast: float
            A number between 0 and 1.  Note that upon initialization the
            colormap has a default contrast value of 0.5.

        Returns: void
        """
        self._contrast = contrast

        self.x_spread = 2 * (1.0 - contrast)
        self.y_spread = 2.0 - 2 * (1.0 - contrast)

        self._build_cdict()

    def update_contrast(self, contrast_diff):
        self.set_contrast(self._contrast + contrast_diff)

    def _clip(self, value):
        """Clip to range 0 to 1"""
        return clip(value, 0, 1)

    def as_mpl_cmap(self):
        return LinearSegmentedColormap("CustomGrayscale", self.cdict)


def clip(value, min_val, max_val):
    """
    Clip a value to a certain range.

    Args:
      value: number
        The value to be clipped.
      min_val: number
        The lowest allowable value.
      max_val: number
        The maximum allowable value.

    Returns:
      clipped: number
        The original value if it is between min_val and max_val.
        If it is less than min_val, return min_val.  If it is greater
        than max_val, return max_val.

    Examples;
      clip(0.1, 0, 1) -> 0.1
      clip(-0.1, 0, 1) -> 0
    """
    return max(min(value, max_val), min_val)



__author__ = "David Rusk <drusk@uvic.ca>"

import wx


class NavPanel(wx.Panel):
    NEXT_LABEL = "Next"
    PREV_LABEL = "Previous"

    def __init__(self, parent, controller):
        super(NavPanel, self).__init__(parent)

        self.controller = controller

        self._init_ui()
        self._bind_events()

    def _init_ui(self):
        self.sbox = wx.StaticBox(self, label="Source Observations")

        self.status_text = wx.StaticText(self, label="Loading...")
        self.next_button = wx.Button(self, wx.ID_FORWARD, label=self.NEXT_LABEL)
        self.prev_button = wx.Button(self, wx.ID_BACKWARD, label=self.PREV_LABEL)

        self._do_layout()

    def _do_layout(self):
        sbox_sizer = wx.StaticBoxSizer(self.sbox, wx.VERTICAL)

        vsizer = wx.BoxSizer(wx.VERTICAL)
        vsizer.Add(self.status_text, flag=wx.ALIGN_CENTER | wx.TOP, border=10)

        hsizer = wx.BoxSizer(wx.HORIZONTAL)
        button_border = 10
        hsizer.Add(self.prev_button, proportion=0, flag=wx.ALL, border=button_border)
        hsizer.Add(self.next_button, proportion=0, flag=wx.ALL, border=button_border)

        vsizer.Add(hsizer)

        sbox_sizer.Add(vsizer, flag=wx.ALIGN_CENTER)

        # Add a bit of border around the box sizer
        border_sizer = wx.BoxSizer(wx.VERTICAL)
        border_sizer.Add(sbox_sizer, flag=wx.EXPAND | wx.ALL, border=10)

        self.SetSizer(border_sizer)

    def _bind_events(self):
        self.next_button.Bind(wx.EVT_BUTTON, self._on_next)
        self.prev_button.Bind(wx.EVT_BUTTON, self._on_prev)

    def _on_next(self, event):
        self.controller.on_next_obs()

    def _on_prev(self, event):
        self.controller.on_previous_obs()

    def set_status(self, current_obs, total_obs):
        assert 0 <= current_obs <= total_obs
        self.status_text.SetLabel("%s of %s" % (current_obs, total_obs))

        # Re-center text
        self.Layout()



# horizons.py
# Query JPL Horizons via the web service
# Wes Fraser, 2010
# rewritten and documented by Michele Bannister, Dec 2010
# rewritten again by MB, Jan 2015

import urllib2 as url
import time
import io

import pandas as pd


'''
Ephemeris parameters:
1. Astrometric RA & DEC  15. Sun sub-long & sub-lat  29. Constellation ID
 *2. Apparent RA & DEC     16. Sub Sun Pos. Ang & Dis  30. Delta-T (CT - UT)
  3.   Rates; RA & DEC     17. N. Pole Pos. Ang & Dis *31. Obs eclip. lon & lat
 *4. Apparent AZ & EL      18. Helio eclip. lon & lat  32. North pole RA & DEC
  5.   Rates; AZ & EL      19. Helio range & rng rate  33. Galactic latitude
  6. Sat. X & Y, pos. ang  20. Obsrv range & rng rate  34. Local app. SOLAR time
  7. Local app. sid. time  21. One-Way Light-Time      35. Earth->Site lt-time
  8. Airmass               22. Speed wrt Sun & obsrvr >36. RA & DEC uncertainty
  9. Vis mag. & Surf Brt   23. Sun-Obs-Targ ELONG ang >37. POS error ellipse
 10. Illuminated fraction  24. Sun-Targ-Obs PHASE ang >38. POS uncertainty (RSS)
 11. Defect of illumin.    25. Targ-Obsrv-Moon/Illum% >39. Range & Rng-rate sig.
 12. Sat. angle separ/vis  26. Obs-Primary-Targ angl  >40. Doppler/delay sigmas
 13. Target angular diam.  27. Pos. Ang;radius & -vel  41. True anomaly angle
 14. Obs sub-lng & sub-lat 28. Orbit plane angle       42. Local app. hour angle
'''

# PARAMS = OrderedDict([('RA-DEC', 1),                    # Units: HMS (HH MM SS.ff) and DMS (DD MM SS.f)
# ('RA_rate-DEC_rate', 3),          # d(RA)/dt is multiplied by the cosine of the declination.
# # Units: ARCSECONDS PER HOUR
#                       ('v_mag-surface_brightness', 9),  # Units: None and VISUAL MAGNITUDES PER SQUARE ARCSECOND
#                       ('r_helio-r_helio_rate', 19),
#                       ('RA_uncert-DEC_uncert', 36),     # RA and DEc uncertainties are 3-sigma. Units: ARCSECONDS
# ])

def parse_orbital_elements(urlData):
    # Get required information back out of the page Horizons returns
    # Extract the orbital elements and epoch of origin - works as long as the name parses, even if the ephemeris
    # request fails
    epoch_possible = False
    for i in range(len(urlData)):
        S = urlData[i].split()
        if len(S) > 0:
            if S[0] == 'EPOCH=':  # moons don't have epochs so this can't be required
                epoch_possible = True
                epochStr = "/".join(S[3].split('-'))
                S = urlData[i + 1].split('=')
                e = float(S[1].split(' QR')[0])
                S = urlData[i + 2].split('=')
                Omega = float(S[1].strip(' W'))
                W = float(S[2].strip(' IN'))
                inc = float(S[3])
                S = urlData[i + 3].split('=')
                a = float(S[1].strip(' MA'))
                M = float(S[2].strip(' ADIST'))
                break

    if epoch_possible:
        return {'a': a, 'e': e, 'i': inc, 'Omega': Omega, 'W': W, 'M': M, 'Epoch': epochStr}
    else:
        return None


# Run a Horizons query
# eg. output = batch("Haumea", "2010-12-28 10:00", "2010-12-29 10:00", 1, su='d')

def batch(object, t, T, step,
          su='d',
          location=None,
          params=[1, 3, 9, 19, 36]):
    if step == None:  # default
        step = 1
    else:
        step = int(step)

    # Construct the query url
    s = "'"
    if not params:
        for i in range(1, 40):  # There are 40 possible pieces of info that Horizons can give back
            s += str(i) + ','
            s += "40'"  # python leaves one off the end in range()
    else:
        for p in params:
            s += "{},".format(p)

    # The pieces of the url that Horizons needs for its processing instructions.
    urlArr = ["http://ssd.jpl.nasa.gov/horizons_batch.cgi?batch=1&COMMAND=",
              '',
              "&MAKE_EPHEM='YES'&TABLE_TYPE='OBSERVER'&START_TIME=",
              '',
              "&STOP_TIME=",
              '',
              "&STEP_SIZE=",
              '',
              "&QUANTITIES=" + s,
              "&CSV_FORMAT='YES'"]

    # Break the object name, start & end dates and the timestep up into appropriate url-formatting
    url_style_output = []
    for obj in [object, t, T]:
        os = obj.split()
        if len(os) > 1:
            ob = "'" + os[0] + '%20' + os[1] + "'"
        else:
            ob = "'" + object + "'"
        url_style_output.append(ob)
    step = "'" + str(step) + "%20" + su + "'"

    # url components
    urlArr[1] = url_style_output[0]  # formatted object name
    urlArr[3] = url_style_output[1]  # start time
    urlArr[5] = url_style_output[2]  # end time
    urlArr[7] = step  # timestep

    urlStr = "".join(urlArr)  # create the url to pass to Horizons

    # Query Horizons; if it's busy, wait and try again in a minute
    done = 0
    while not done:
        urlHan = url.urlopen(urlStr)
        urlData = urlHan.readlines()
        urlHan.close()
        if len(urlData[0].split()) > 1:
            if "BUSY:" <> urlData[0].split()[1]:
                done = 1
            else:
                print urlData[0],
                print "Sleeping 60 s and trying again"
                time.sleep(60)
        else:
            done = 1

    EPHEM_CSV_START_MARKER = '$$SOE'
    EPHEM_CSV_END_MARKER = '$$EOE'
    ephemCSV_start = None
    ephemCSV_end = None
    for i, dataLine in enumerate(urlData):
        if dataLine.strip() == EPHEM_CSV_START_MARKER:
            ephemCSV_start = i
        elif dataLine.strip() == EPHEM_CSV_END_MARKER:
            ephemCSV_end = i
    assert ephemCSV_start is not None, 'No ephem start'
    assert ephemCSV_end is not None, 'No ephem end'
    assert ephemCSV_start < ephemCSV_end, 'Ephem are a bit odd'
    # The header, the lines *after* the start marker, up to (but not including, because it is a slice) the end marker
    csv_lines = [urlData[ephemCSV_start - 2]] + urlData[ephemCSV_start + 1: ephemCSV_end]
    ephemCSV = ''.join(csv_lines)
    ephemCSVfile = io.BytesIO(ephemCSV)
    ephemerides = pd.DataFrame.from_csv(ephemCSVfile)

    orbital_elements = parse_orbital_elements(urlData)

    return orbital_elements, ephemerides

# Run the script from the command line
if __name__ == "__main__":
    elems, ephems = batch("Ceres", "2010-12-28 10:00", "2010-12-30 10:00", 1, su='d')  # for example

    # just a few test print statements
    print(elems)  # elements
    print(ephems.keys())



import ctypes
import os
import tempfile
from astropy import coordinates
from astropy import units
from ossos.mpc import Time, Observation

LIBORBFIT='/usr/local/lib/liborbfit.so'

__author__ = 'jjk'

class Orbfit(object):

    def __init__(self, observations=[], abg=[], residuals=[]):
        self.orbfit = ctypes.CDLL(LIBORBFIT)
        self.abg = abg
        self.residuals = residuals
        self.observations = observations
        self._fit_radec()

    def _fit_radec(self):
        """
        call fit_radec of BK passing in the observations.

        """

        # call fitradec with mpcfile, abgfile, resfile
        self.orbfit.fitradec.restype = ctypes.c_int
        self.orbfit.fitradec.argtypes = [ ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p ]

        mpc_file = tempfile.NamedTemporaryFile(suffix='.mpc')
        for observation in self.observations:
            mpc_file.write("{}\n".format(str(observation)))
        mpc_file.seek(0)

        abg_file = tempfile.NamedTemporaryFile()
        res_file = tempfile.NamedTemporaryFile()

        self.orbfit.fitradec(ctypes.c_char_p(mpc_file.name),
                             ctypes.c_char_p(abg_file.name),
                             ctypes.c_char_p(res_file.name))

        self.abg = abg_file
        self.abg.seek(0)
        self.residuals = res_file
        self.residuals.seek(0)

    def predict(self, date, obs_code=568):
        """
        use the bk predict method to compute the location of the source on the given date.
        """
        time = Time(date, scale='utc', precision=6)
        jd = ctypes.c_double(time.jd)
        # call predict with agbfile, jdate, obscode
        self.orbfit.predict.restype = ctypes.POINTER(ctypes.c_double * 5)
        self.orbfit.predict.argtypes = [ ctypes.c_char_p, ctypes.c_double, ctypes.c_int ]
        predict = self.orbfit.predict(ctypes.c_char_p(self.abg.name),
                       jd,
                       ctypes.c_int(obs_code))
        self.coordinate = coordinates.SkyCoord(predict.contents[0],
                                               predict.contents[1],
                                               unit=(units.degree, units.degree))
        self.dra = predict.contents[2]
        self.ddec = predict.contents[3]
        self.pa = predict.contents[4]
        self.date = str(time)


if __name__ == '__main__':

    mpc_lines=("     HL7j2    C2013 04 03.62926 17 12 01.16 +04 13 33.3          24.1 R      568",
               "     HL7j2    C2013 04 04.58296 17 11 59.80 +04 14 05.5          24.0 R      568",
               "     HL7j2    C2013 05 03.52252 17 10 38.28 +04 28 00.9          23.4 R      568",
               "     HL7j2    C2013 05 08.56725 17 10 17.39 +04 29 47.8          23.4 R      568")

    observations = []
    for line in mpc_lines:
        observations.append(Observation().from_string(line))

    os.environ['ORBIT_EPHEMERIS']='/Users/jjk/MOP/config/binEphem.405'
    os.environ['ORBIT_OBSERVATORIES']='/Users/jjk/MOP/config/observatories.dat'
    HL7j2 = Orbfit(observations=observations)
    for observation in observations:
        HL7j2.predict(observation.date)
        dra = coordinates.Angle(HL7j2.coordinate.ra - observation.coordinate.ra)
        if dra.degrees > 180 :
            dra = dra - coordinates.Angle(360, unit=units.degree)
        ddec = coordinates.Angle(HL7j2.coordinate.dec - observation.coordinate.dec)
        if ddec.degrees > 180:
            dra = ddec - coordinates.Angle(360, unit=units.degree)

        print "Input  : {} {} {} [ {:+4.2f} {:+4.2f} ]".format(observation.date,
                                               observation.ra,
                                               observation.dec,
                                               dra.degrees*3600.0,
                                               ddec.degrees*3600.0)
        #print "Predict: {} {} {}".format(HL7j2.date,
        #                                 HL7j2.coordinate.ra.format(unit=units.hour,
        #                                                           precision=3,
        #                                                            pad=True,
        #                                                            sep=" "),
        #                                 HL7j2.coordinate.dec.format(unit=units.degree,
        #                                                            decimal=False,
        #                                                            precision=2,
        #                                                            pad=True,
        #                                                            sep=" ",
        #                                                            alwayssign=True))



import traceback
import tempfile

from astropy import units

from astropy.coordinates import SkyCoord

from astropy.io import fits

from ...astrom import SourceReading, Observation
from ...gui import logger
from ... import storage
from ossos.gui import config

__author__ = "David Rusk <drusk@uvic.ca>"


class SourceCutout(object):
    """
    A cutout around a source.
    """

    def __init__(self, reading, hdulist, apcor=None, zmag=None, radius=None):
        """
        :param reading: A source reading giving the measurement of the object associated with this cutout.
        :param hdulist: the HDUList containing the cutout.
        :param apcor: The aperture correction for this cutout.
        :param zmag: The zeropoint of the flux calibration for this cutout.
        :return:
        """
        logger.debug("building a SourceCutout.")
        assert isinstance(reading, SourceReading)
        assert isinstance(hdulist, fits.HDUList)
        # assert isinstance(apcor, ApcorData)

        self.radius = radius
        self.reading = reading
        self.hdulist = hdulist
        self.apcor = apcor
        self.zmag = zmag
        self.pixel_y = self.pixel_x = self._extno = None

        if self.reading.x is None or self.reading.y is None or (self.reading.x == -9999 and self.reading.y == -9999):
            (x, y) = self.hdulist[self.extno].wcs.sky2xy(self.reading.ra, self.reading.dec)
            self.reading.pix_coord = hdulist[self.extno].converter.get_inverse_converter().convert((x, y))
            self.reading.obs.ccdnum = self.extno - 1

        self.original_observed_ext = self.extno
        self.original_observed_x = self.reading.x
        self.original_observed_y = self.reading.y

        self.observed_x = self.original_observed_x
        self.observed_y = self.original_observed_y

        if self.pixel_x is None or self.pixel_y is None:
            self.pixel_x, self.pixel_y = self.get_pixel_coordinates(self.flip_flip(self.observed_source_point))

        self._ra = self.reading.ra * units.degree
        self._dec = self.reading.dec * units.degree
        self._ext = self.original_observed_ext

        self._stale = False
        self._adjusted = False
        self._comparison_image = None
        self._tempfile = None
        self._bad_comparison_images = [self.hdulist[-1].header.get('EXPNUM', None)]
        logger.debug("type X/Y {}/{}".format(type(self.pixel_x),
                                             type(self.pixel_y)))

    @property
    def extno(self):
        if self._extno is None:
            self._extno = SourceCutout.resolve_extno(self.hdulist, self.reading.get_ccd_num())
        return self._extno

    @staticmethod
    def resolve_extno(hdulist, ccdnum):
        for (extno, hdu) in enumerate(hdulist):
            if ccdnum == int(hdu.header.get('EXTVER', -1)):
                return extno
        return len(hdulist) - 1

    def flip_flip(self, point):
        """
        Sometimes we 'flipped' the X/Y coordinates of the image during processing.  This function takes an
        (x,y) location from the processing pipeline and returns the x/y location relative to the original
        reference frame.   (Correctly checks if the coordinate is flipped to begin with.)

        @param point: an x/y location on the image.
        @type point: [x, y]
        @return:  the x/y location on the un-fliped frame
        """
        x, y = point
        if self.reading.compute_inverted():
            naxis1 = self.reading.obs.header.get(self.reading.obs.HEADER_IMG_SIZE_X,
                                                 self.reading.obs.header.get('NAXIS1', 256))
            naxis2 = self.reading.obs.header.get(self.reading.obs.HEADER_IMG_SIZE_Y,
                                                 self.reading.obs.header.get('NAXIS2', 256))
            logger.debug("Got that image has size {}, {}".format(naxis1, naxis2))

            x1 = x
            y1 = y
            x = int(naxis1) - x + 1
            y = int(naxis2) - y + 1
            logger.debug("Inverted source location from {},{} to {},{}".format(x1, y1, x, y))

        return x, y

    @property
    def astrom_header(self):
        return self.hdulist[len(self.hdulist) - 1].header

    @property
    def fits_header(self):
        return self.hdulist[len(self.hdulist) - 1].header

    @property
    def observed_source_point(self):
        return self.observed_x, self.observed_y

    @property
    def pixel_source_point(self):
        return self.pixel_x, self.pixel_y

    def update_pixel_location(self, new_pixel_location, extno=1):
        self.pixel_x, self.pixel_y = new_pixel_location
        self.observed_x, self.observed_y = self.get_observed_coordinates(
            new_pixel_location, extno=extno)
        self._ext = extno
        self._stale = True
        self._adjusted = True

    def reset_source_location(self):
        self.observed_x = self.original_observed_x
        self.observed_y = self.original_observed_y
        self.pixel_x, self.pixel_y = self.get_pixel_coordinates(
            self.observed_source_point)

        self._stale = True
        self._adjusted = False

    @property
    def ra(self):
        self._lazy_refresh()
        return self._ra

    @property
    def dec(self):
        self._lazy_refresh()
        return self._dec

    def is_adjusted(self):
        return self._adjusted

    def get_pixel_coordinates(self, point, extno=None):
        """
        Retrieves the pixel location of a point within the current HDUList given the
        location in the original FITS image.  This takes into account that
        the image may be a cutout of a larger original.

        The approximate RA/DEC are required in-case there is more than one HDU for this source,
        that can happen when a source is near the pixel boundary.

        Args:
          point: tuple(float, float)
            (x, y) in original.

        Returns:
          (x, y) pixel in this image.
        """
        if not extno:
            extno = self.original_observed_ext
        return self.hdulist[extno].converter.convert(point)

    def get_observed_coordinates(self, point, extno=None):
        """
        Retrieves the location of a point using the coordinate system of
        the original observation, i.e. the original image before any
        cutouts were done.

        Args:
          point: tuple(float, float)
            The pixel coordinates.

        Returns:
          (x, y) in the original image coordinate system.
        """
        return self.hdulist[extno].converter.get_inverse_converter().convert(point)

    def pix2world(self, x, y, usepv=True):
        return self.hdulist[self.extno].wcs.xy2sky(x, y, usepv=usepv)

    def world2pix(self, ra, dec, usepv=True):
        """
        Convert a given RA/DEC position to the Extension X/Y location.

         Here we loop over the hdulist to find the extension is RA/DEC are from and then return the appropriate X/Y
        :param ra: The Right Ascension of the point
        :type ra: Quantity
        :param dec: The Declination of the point
        :type dec: Quantity
        :return: X, Y, Extension position of source in cutout reference frame
        :rtype: (float, float, int)
        """
        logger.debug("Converting (ra,dec) to (x,y) given {:8.3f} {:8.3f} for image {}".format(
            ra, dec, self.reading.get_exposure_number()))
        for idx in range(1, len(self.hdulist)):
            logger.debug("Trying convert using extension: {0}".format(idx))
            hdu = self.hdulist[idx]
            x, y = hdu.wcs.sky2xy(ra, dec, usepv=usepv)
            logger.debug("Tried converting RA/DEC {0},{1} to X/Y and got {2},{3}".format(ra, dec, x, y))
            if 0 < x < hdu.header['NAXIS1'] and 0 < y < hdu.header['NAXIS2']:
                logger.debug("Inside the frame.")
                return x, y, idx
        return None, None, None

    def get_observed_magnitude(self):
        # NOTE: this import is only here so that we don't load up IRAF
        # unnecessarily (ex: for candidates processing).
        """
        Get the magnitude at the current pixel x/y location.

        :return: Table
        """
        from ossos import daophot

        max_count = float(self.astrom_header.get("MAXCOUNT", 30000))
        tmp_file = self._hdulist_on_disk()
        if not self.zmag:
            self.zmag = self.hdulist[self._ext].header.get('PHOTZP', 0.0)
        try:
            phot = daophot.phot_mag(tmp_file,
                                    self.pixel_x, self.pixel_y,
                                    aperture=self.apcor.aperture,
                                    sky=self.apcor.sky,
                                    swidth=self.apcor.swidth,
                                    apcor=self.apcor.apcor,
                                    zmag=self.zmag,
                                    maxcount=max_count, extno=1)
            if not self.apcor.valid:
                phot['PIER'][0] = 1
            return phot
        finally:
            self.close()

    def _hdulist_on_disk(self):
        """
        IRAF routines such as daophot need input on disk.

        Returns:
          filename: str
            The name of the file containing the FITS data.
        """
        if self._tempfile is None:
            self._tempfile = tempfile.NamedTemporaryFile(
                mode="r+b", suffix=".fits")
            self.hdulist[self._ext].writeto(self._tempfile.name)
        return self._tempfile.name

    def close(self):
        """
        Once we are done with the on disk content we should close the filehandle.
        """
        if self._tempfile is not None:
            self._tempfile.close()
            self._tempfile = None

    def _lazy_refresh(self):
        if self._stale:
            self._update_ra_dec()
            self._stale = False

    def _update_ra_dec(self):
        hdu = self.hdulist[self.original_observed_ext]
        self._ra, self._dec = hdu.wcs.xy2sky(self.pixel_x, self.pixel_y)

    @property
    def comparison_image(self):
        if self._comparison_image is None:
            self.retrieve_comparison_image()
        return self._comparison_image

    def retrieve_comparison_image(self):
        """
        Search the DB for a comparison image for this cutout.
        """
        # selecting comparator when on a comparator should load a new one.

        ref_ra = self.reading.ra * units.degree
        ref_dec = self.reading.dec * units.degree
        radius = self.radius is not None and self.radius or config.read('CUTOUTS.SINGLETS.RADIUS') * units.arcminute
        query_result = storage.cone_search(ref_ra, ref_dec, radius, radius)  # returns an astropy.table.table.Table
        comparison = None
        if len(query_result['collectionID']) > 0:  # are there any comparison images even available on that sky?
            for collectionID in query_result['collectionID']:
                if collectionID not in self._bad_comparison_images:
                    comparison = collectionID
                    self._bad_comparison_images.append(comparison)
                    try:
                        hdu_list = storage.ra_dec_cutout(storage.dbimages_uri(comparison),
                                                         SkyCoord(ref_ra, ref_dec), radius)
                        obs = Observation(str(comparison), 'p',
                                          ccdnum=int(hdu_list[-1].header.get('EXTVER', 0)))
                        x = hdu_list[-1].header.get('NAXIS1', 0) // 2.0
                        y = hdu_list[-1].header.get('NAXIS2', 0) // 2.0
                        reading = SourceReading(x, y, self.reading.x, self.reading.y,
                                                ref_ra, ref_dec, self.reading.x, self.reading.y, obs)
                        self._comparison_image = SourceCutout(reading, hdu_list)
                    except Exception as ex:
                        logger.error("{} {}".format(type(ex), str(ex)))
                        logger.error(traceback.format_exc())
                        continue
                    break
            if comparison is None:
                logger.critical(str(self.fits_header))
                self._comparison_image = None
                return
        else:
            query_result.pprint()
            logger.info("No comparison images available for this piece of sky.")
            print "No comparison images available for this piece of sky."
            self._comparison_image = None
            return



from astropy import units
from astropy.units import Quantity
from ossos.gui import config

from ..core import Downloader, ApcorData
from ..cutouts.source import SourceCutout
from ... import storage
from ...astrom import SourceReading
from ...gui import logger


class ImageDownloader(Downloader):

    def download(self, reading, needs_apcor=False):

        hdulist = storage.get_hdu(reading.get_image_uri())
        apcor = needs_apcor and self.download_apcor(reading.get_apcor_uri()) or None
        zmag = self.download_zmag(reading.get_zmag_uri())
        return SourceCutout(reading, hdulist, apcor, zmag=zmag)


class ImageCutoutDownloader(Downloader):
    """
    Downloads a slice of an image relevant to examining a (potential) source.
    """

    def __init__(self, slice_rows=500, slice_cols=500, vosclient=None):
        """
        Constructor.

        Args:
          slice_rows, slice_cols: int
            The number of rows and columns (pixels) to slice out around the
            source.  Leave as None to use default configuration values.
        """
        super(ImageCutoutDownloader, self).__init__()

        # self.cutout_calculator = CutoutCalculator(slice_rows, slice_cols)

    def download_cutout(self, reading, focus=None, needs_apcor=False):
        """
        Downloads a cutout of the FITS image for a given source reading.

        Args:
          reading: ossos.astrom.SourceReading
            The reading which will be the focus of the downloaded image.
          focus: tuple(int, int)
            The x, y coordinates that should be the focus of the downloaded
            image.  These coordinates should be in terms of the
            source_reading parameter's coordinate system.
            Default value is None, in which case the source reading's x, y
            position is used as the focus.
          needs_apcor: bool
            If True, the apcor file with data needed for photometry
            calculations is downloaded in addition to the image.
            Defaults to False.

        Returns:
          cutout: ossos.downloads.data.SourceCutout
        """
        logger.debug("Doing download_cutout with inputs: reading:{} focus:{} needs_apcor:{}".format(reading,
                                                                                                    focus,
                                                                                                    needs_apcor))
        assert isinstance(reading, SourceReading)
        min_radius = config.read('CUTOUTS.SINGLETS.RADIUS')
        if not isinstance(min_radius, Quantity):
            min_radius = min_radius * units.arcminute
        radius = max(reading.uncertainty_ellipse.a,
                     reading.uncertainty_ellipse.b) * 2.5 + min_radius

        logger.debug("got radius for cutout: {}".format(radius))
        image_uri = reading.get_image_uri()
        logger.debug("Getting cutout at {} for {}".format(reading.reference_sky_coord, image_uri))
        hdulist = storage.ra_dec_cutout(image_uri, reading.reference_sky_coord, radius)
        logger.debug("Getting the aperture correction.")
        apcor = ApcorData.from_string("4 15 1.0 -99")

        if needs_apcor:
            try:
                apcor = self.download_apcor(reading.get_apcor_uri())
            except Exception as ex:
                pass

        zmag = None
        try:
            zmag = self.download_zmag(reading.get_zmag_uri())
        except Exception as ex:
            pass

        header = reading.obs.header
        logger.debug("Sending back the source reading.")
        return SourceCutout(reading, hdulist, apcor, zmag=zmag, radius=radius)



__author__ = "David Rusk <drusk@uvic.ca>"

import wx

from ...gui import config, logger
from ...gui.views.listctrls import ListCtrlPanel
from ...gui.views.navigation import NavPanel
from ...gui.views.validation import SourceValidationPanel, MPCPanel


class MainFrame(wx.Frame):
    """
    This is the main window of the application.  It should not be
    manipulated externally (it should be considered an implementation detail
    of the ApplicationView).  Therefore, updates should come from the
    ApplicationView.
    """

    def __init__(self, controller, track_mode=False):
        logger.debug("Building mainframe.")
        size = (config.read("UI.DIMENSIONS.WIDTH"),
                config.read("UI.DIMENSIONS.HEIGHT"))
        super(MainFrame, self).__init__(None, title="Moving Object Pipeline",
                                        size=size)

        self.track_mode = track_mode
        self.controller = controller

        self._init_ui_components()

        # needed for keybinds to work on startup
        self.main_panel.SetFocus()

    def _init_ui_components(self):
        self.main_panel = _FocusablePanel(self, style=wx.RAISED_BORDER)
        self.control_panel = wx.Panel(self.main_panel)
        self.main_panel.use_as_focus(self.control_panel)

        self.nav_view = NavPanel(self.control_panel, self.controller)

        self.data_view = self._create_data_notebook()

        self.validation_view = SourceValidationPanel(self.control_panel, self.controller)

        if self.track_mode:
            self.mpc_save_view = MPCPanel(self.control_panel, self.controller)

        self._do_layout()

    def _do_layout(self):
        control_sizer = wx.BoxSizer(wx.VERTICAL)
        control_sizer.Add(self.nav_view, 1, flag=wx.EXPAND)
        control_sizer.Add(self.data_view, 2, flag=wx.EXPAND)
        control_sizer.Add(self.validation_view, 1, flag=wx.EXPAND)

        if self.track_mode:
            control_sizer.Add(self.mpc_save_view, 1, flag=wx.EXPAND)

        self.control_panel.SetSizerAndFit(control_sizer)

        main_sizer = wx.BoxSizer(wx.HORIZONTAL)
        main_sizer.Add(self.control_panel, flag=wx.EXPAND)

        self.main_panel.SetSizerAndFit(main_sizer)

        self.main_sizer = main_sizer

    def _create_data_notebook(self):
        notebook = wx.Notebook(self.control_panel)

        columns = ("{:15s}".format("Key".center(10)),
                    "{:15s}".format("Value".center(10)))
        self.reading_data_panel = ListCtrlPanel(notebook, columns)
        self.obs_header_panel = ListCtrlPanel(notebook, columns)

        notebook.AddPage(self.reading_data_panel, "Readings")
        notebook.AddPage(self.obs_header_panel, "Observation Header")

        return notebook

    def update_displayed_data(self, reading_data, header_data_list):
        self.reading_data_panel.populate_list(reading_data)
        self.obs_header_panel.populate_list(header_data_list)

    def set_observation_status(self, current_obs, total_obs):
        self.nav_view.set_status(current_obs, total_obs)

    def disable_validation(self):
        self.validation_view.disable()

    def enable_validation(self):
        self.validation_view.enable()

    def is_source_validation_enabled(self):
        return self.validation_view.is_validation_enabled()

    def add_to_main_sizer(self, widget):
        self.main_sizer.Add(widget, flag=wx.EXPAND)
        self.main_sizer.Layout()


class _FocusablePanel(wx.Panel):
    """
    Work-around used to make sure the right windows have focus so the
    keybind accelerator table works.
    """

    def __init__(self, *args, **kwargs):
        super(_FocusablePanel, self).__init__(*args, **kwargs)

        self._focus = None

    def use_as_focus(self, widget):
        self._focus = widget

    def SetFocus(self, **kwargs):
        """
        Over-rides normal behaviour of shifting focus to any child.  Prefers
        the one set explicityly by use_as_focus.
        :param **kwargs:
        """
        if self._focus is not None:
            self._focus.SetFocus()
        else:
            # fall back on the default behaviour
            super(_FocusablePanel, self).SetFocus()



__author__ = 'Michele Bannister and Jean-Marc Petit'

import os

import matplotlib.pyplot as plt
from astropy.table import Table
import numpy as np

from palettable.colorbrewer import sequential
from palettable import tableau
import prettyplotlib as ppl
import plot_fanciness
from utils import read_smooth_fit, square_fit

path = '/Users/bannisterm/Dropbox/Papers in progress/OSSOS/First_quarter/data/'

def plot_smooth_fit(i, block, ax, colours, pwd, offset=0, single=False):
    characterisation = {'13AE': 24.09, '13AO': 24.40}
    ls = ['-', '--', ':']

    smooth_parameter_files = filter(lambda name: name.startswith('smooth'), os.listdir('{}/{}/'.format(pwd, block)))
    smooth_parameter_files.sort(key=lambda x: float(x.split('-')[1]))
    x = np.arange(20.5, 25.1, 0.01)
    if single:  # just do the first one, the motion rate targeted by the main survey objectives
        smooth_parameter_files = smooth_parameter_files[0:1]
    for j, fn in enumerate(smooth_parameter_files):
        pd, ps, mag_limit = read_smooth_fit('{}/{}/{}'.format(pwd, block, fn))
        ys = square_fit(x, ps)

        print block, fn, mag_limit, square_fit(mag_limit, ps)

        if single:  # only want to add label to the legend if there is only one given
            ax[i].plot(x + (offset * j), ys * 100.,
                       ls=ls[j], color=colours[j],
                       label='smooth fit {}-{} "/hr overall'.format(fn.split('-')[1], fn.split('-')[2]))
        else:
            ax[i].plot(x + (offset * j), ys * 100.,
                       ls=ls[j], color=colours[j])

        # last, add the vertical line for characterisation limit for the block
        if j == len(smooth_parameter_files) - 1:
            # defined on target KBO population, modulo tracking efficiency & other factors: not on 40% mag limit
            ax[i].vlines(characterisation[block], 0., 90,
                         linestyles=':', alpha=0.7)
            ax[i].annotate("$m_{{characterized}}$ = {:.2f}".format(characterisation[block]),
                           (characterisation[block], 90), size=10, color='k', ha='center')


def plot_eff_data(i, block, ax, colours, pwd, offset):
    fmt = ['^', 'x', 'o']
    filenames = filter(lambda name: name.endswith('mag-rate.eff'), os.listdir('{}/{}/'.format(pwd, block)))
    filenames.sort(key=lambda x: float(x.split('-')[0]))
    for j, fn in enumerate(filenames):
        fn_pwd = '{}/{}/{}'.format(pwd, block, fn)
        eff = Table.read(fn_pwd, format='ascii', guess=False, delimiter=' ', data_start=0, comment='#',
                         names=['mag', 'eff', 'num_planted', 'd_eff_plus', 'd_eff_minus'],
                         header_start=None)
        ax[i].errorbar(eff['mag'] + (offset * j), eff['eff'] * 100., yerr=eff['d_eff_plus'] * 100.,
                       fmt=fmt[j], capsize=1, elinewidth=0.5, ms=3.5,
                       label='{} "/hr'.format(fn.split('_')[0]),
                       mfc=colours[j], mec=colours[j], ecolor=colours[j])  # enforce: colour_cycle didn't set all


def plot_eff_by_rate_of_motion(ax, blocks):
    pwd = path + 'efficiency_motion_rates'
    outfile = 'OSSOS_e-o_efficiency.pdf'
    colours = sequential.Greens_6.mpl_colors[::-1]  # want efficiency for most-targeted <8"/hr objects most prominent
    colours = colours[0:2] + [colours[3]]
    offset = 0.017  # set the error bars of data off from each other slightly for legibility; also displace smooth fits

    for i, block in enumerate(blocks):
        plot_smooth_fit(i, block, ax, colours, pwd, offset=offset)
        plot_eff_data(i, block, ax, colours, pwd, offset)
        ax[i].grid(True, alpha=0.3)
        ax[i].set_ylabel('efficiency (%)')
        plot_fanciness.remove_border(ax[i])
        ppl.legend(ax[i], loc='lower left', title=block, numpoints=1, fontsize='small', handletextpad=0.5)

    return outfile


def plot_eff_by_user(ax, blocks):
    pwd = path + 'efficiency_personal_bias/'
    outfile = 'e-o_efficiency_by_user.pdf'
    col = tableau.ColorBlind_10.mpl_colors
    col = col[0:5] + [col[8]]

    # no. files examined, from plots Charles made: see http://wiki.ossos-survey.org/index.php/Core_Teleconf_2014_10_21
    examined = {'13AE': {'jjk': 163, 'jkavelaars': 109, 'mtb55': 418},
                '13AO': {'jkavelaars': 53, 'mtb55': 71, 'bgladman': 111, 'montys': 440, 'ptsws': 74}}
    # swap out user ID for anonymous 'pnum' values in the final displayed plot
    user_blindness = {'mtb55': 'p1',
                      'jjk': 'p2',
                      'jkavelaars': 'p3',
                      'montys': 'p4',
                      'bgladman': 'p5',
                      'ptsws': 'p6'
                      }
    # Line colour and symbol consistent between plots for participant.
    colours = {'jjk': col[2], 'jkavelaars': col[1], 'mtb55': col[0], 'bgladman': col[3], 'montys': col[4],
               'ptsws': col[5]}
    markers = {'jjk': '^', 'jkavelaars': 'd', 'mtb55': '*', 'bgladman': 'o', 'montys': '.', 'ptsws': 'x'}

    for i, block in enumerate(blocks):
        user_eff_files = filter(lambda x: x.__contains__(block), os.listdir(pwd))
        # user_eff_files.sort(key=lambda x: examined[block][x.split('.')[2]], reverse=True)  # biggest to smallest
        user_eff_files.sort(key=lambda x: user_blindness[x.split('.')[2]])
        for j, fn in enumerate(user_eff_files):
            eff = Table.read(pwd + fn, names=['mag', 'eff'], format='ascii')
            user = fn.split('.')[2]
            ax.plot(eff['mag'], eff['eff'] * 100,
                       c=colours[user], marker=markers[user], ms=5, mec=colours[user], alpha=0.7,
                       # line thickness scaled by number of files examined: most first so that thickest line at the back
                       linewidth=examined[block][user] * 0.02,
                    label="{}: {}".format(user_blindness[user], examined[block][user]),
                    zorder=1 - examined[block][user])
        # not the appropriate smooth fit to be adding here.
        # plot_smooth_fit(i, block, ax, ['k'], path + 'efficiency_motion_rates', single=True)

        ax.grid(True, alpha=0.3)
        ax.set_ylabel('efficiency (%)')
        ax.set_ylim([0., 100.])
        plot_fanciness.remove_border(ax)
        ppl.legend(ax, loc='lower left', title=block, numpoints=1, fontsize='small', handletextpad=0.5)

    return outfile


if __name__ == '__main__':
    fig, ax = plt.subplots(nrows=2, ncols=1, sharex=True, figsize=(9, 4))
    fig.subplots_adjust(hspace=0.1)
    blocks =  ['13AE', '13AO'] # [',['13AO']  #

    outfile = plot_eff_by_rate_of_motion(ax, blocks)
    # outfile = plot_eff_by_user(ax, blocks)

    plt.xlabel("$m_{r}$")
    plt.xlim([21.1, 25.])
    plt.draw()
    plt.savefig(outfile, transparent=True, bbox_inches='tight')



from unittest import TestCase

__author__ = 'jjk'

import util

class TestGet_pixel_bounds_from_datasec_keyword(TestCase):

    def test_get_pixel_bounds_from_datasec_keyword(self):
        """Check that correct tuple is returned"""
        self.assertEquals(util.get_pixel_bounds_from_datasec_keyword("[2:2040,2:4009]"), ((2, 2040), (2, 4009)))



from astropy import units
from astropy.units import Quantity

from ..downloads.cutouts.source import SourceCutout
from ..gui.models.transactions import TransAckValidationModel
from ..gui.models.workload import TracksWorkUnit
from ..gui.views.appview import ApplicationView

__author__ = "David Rusk <drusk@uvic.ca>"
import math
from ..downloads.core import Downloader
from ..gui.autoplay import AutoplayManager
from ..gui import config, logger
from ..gui import events
from ..gui.models.exceptions import (ImageNotLoadedException,
                                     NoWorkUnitException)
from .. import mpc
from ..orbfit import OrbfitError, Orbfit


class AbstractController(object):
    """
    The class that handles interactions between the data model and the users view of that data.
    """
    def __init__(self, model, view):
        """

        @param model: The model that handles getting the data.
        @type model: TransAckValidationModel
        @param view: the display that is showing the controller interaction
        @type view: ApplicationView
        """
        self.model = model
        self.view = view
        events.subscribe(events.CHANGE_IMAGE, self.on_change_image)
        events.subscribe(events.IMG_LOADED, self.on_image_loaded)
        events.subscribe(events.NO_AVAILABLE_WORK, self.on_no_available_work)
        self.mark_prediction = False
        self.mark_source = True
        self.use_pixel_coords = True
        self.autoplay_manager = AutoplayManager(model)
        self.downloader = Downloader()
        self.image_loading_dialog_manager = ImageLoadingDialogManager(view)
        self.align = True

    def get_view(self):
        """
        @rtype: ApplicationView
        """
        return self.view

    def place_marker(self, x, y, radius=10, colour='r'):
        self.view.place_marker(self.model.get_current_cutout(), x, y, radius, colour=colour)

    def display_current_image(self):
        logger.debug("Displaying image.")
        try:
            cutout = self.model.get_current_cutout()
            source = self.model.get_current_source()
            reading = self.model.get_current_reading()
            self.view.image_viewer.mark_source = self.mark_source
            self.view.image_viewer.mark_prediction = self.mark_prediction
            self.view.display(cutout, self.use_pixel_coords)
            logger.debug("Align: {}".format(self.align))
            if self.align:
                self.view.align(cutout, reading, source)
            else:
                self.view.image_viewer.ds9.set("wcs align yes")
        except ImageNotLoadedException as ex:
            logger.info("Waiting to load image: {}".format(ex))
            self.image_loading_dialog_manager.wait_for_item(ex.requested_item)
            pass
        except NoWorkUnitException as ex:
            logger.debug("No work? {}".format(ex))
            return False
        except Exception as ex:
            logger.error("Exception in get_current_cutout: {}".format(ex))
            return False

        logger.debug("Displaying metadata.")
        try:
            self.view.update_displayed_data(self.model.get_reading_data(),
                                            self.model.get_header_data_list())
        except Exception as ex:
            logger.error(type(ex))
            logger.error(str(ex))
            logger.error("Failed to get header for {}".format(self.model.get_current_reading()))
            pass

        self.view.set_observation_status(
            self.model.get_current_obs_number() + 1,
            self.model.get_obs_count())

        self.model.acknowledge_image_displayed()

        return True

    def on_reposition_source(self, new_x, new_y):
        try:
            self.model.update_current_source_location((new_x, new_y))
        except ImageNotLoadedException:
            pass

    def on_image_loaded(self, event):
        displayable_item = event.data
        self.image_loading_dialog_manager.set_item_done(displayable_item)

        if displayable_item == self.model.get_current_displayable_item():
            self.display_current_image()

    def on_change_image(self, event):
        logger.debug("Change Image Event: {}".format(event))
        if self.model.is_current_item_processed():
            self.view.disable_source_validation()
        else:
            self.view.enable_source_validation()

        self.display_current_image()

    def on_no_available_work(self, event):
        logger.debug("Available work event: {}".format(event))
        self.view.hide_image_loading_dialog()

        if self.model.get_num_items_processed() == 0:
            self.view.show_empty_workload_dialog(
                self.model.get_working_directory())
            self._do_exit()
        else:
            should_exit = self.view.all_processed_should_exit_prompt()
            if should_exit:
                self._do_exit()

    def on_use_singlet_view(self):
        self.model.use_singlets()
        self.view.use_singlets()
        self.display_current_image()

    def on_use_triplet_view(self):
        self.model.use_triplets()
        self.view.use_triplets()
        self.display_current_image()

    def on_enable_auto_sync(self):
        self.model.enable_synchronization()

    def on_disable_auto_sync(self):
        self.model.disable_synchronization()

    def on_enable_autoplay(self):
        self.autoplay_manager.start_autoplay()

    def on_disable_autoplay(self):
        self.autoplay_manager.stop_autoplay()

    def on_toggle_autoplay_key(self):
        """
        The user has pressed the keybind for toggling autoplay.
        """
        if self.autoplay_manager.is_running():
            self.autoplay_manager.stop_autoplay()
            self.view.set_autoplay(False)
        else:
            self.autoplay_manager.start_autoplay()
            self.view.set_autoplay(True)

    def on_toggle_reticule_key(self):
        self.view.toggle_reticule()

    def on_show_keymappings(self):
        self.view.show_keymappings()

    def on_exit(self):
        self._do_exit()

    def _do_exit(self):
        self.autoplay_manager.stop_autoplay()
        self.view.close()
        self.model.exit()

    def on_load_comparison(self, research=False):
        raise NotImplementedError()

    def on_next_obs(self):
        self.model.next_obs()

    def on_previous_obs(self):
        self.model.previous_obs()

    def on_accept(self):
        raise NotImplementedError()

    def on_reject(self):
        raise NotImplementedError()

    def on_reset_source_location(self):
        try:
            self.model.reset_current_source_location()
            self.view.refresh_markers()
        except ImageNotLoadedException:
            pass

    def on_toggle_align(self):
        self.align = not self.align
        print self.align and "DS9 Focus Follow On" or "DS9 Focus Follow Off"
        logger.debug("Setting align to : {}".format(self.align))

    def on_reset_colormap(self):
        self.view.reset_colormap()


class ProcessRealsController(AbstractController):
    """
    The main controller of the process reals task.  Sets up the view and
    handles user interactions.
    """

    def __init__(self, model, view, name_generator):
        assert isinstance(model, TransAckValidationModel)
        assert isinstance(view, ApplicationView)
        super(ProcessRealsController, self).__init__(model, view)
        self.name_generator = name_generator
        self.is_discovery = True

    def _generate_provisional_name(self):
        return self.name_generator.generate_name(
            self.model.get_current_astrom_header(),
            self.model.get_current_fits_header())

    def on_load_comparison(self, research=False):
        pass

    def on_accept(self, auto=False):
        """
        Initiates acceptance procedure, gathering required data.
        """
        if self.model.is_current_source_named():
            provisional_name = self.model.get_current_source_name()
        else:
            provisional_name = self._generate_provisional_name()

        band = self.model.get_current_band()
        logger.debug("Got band {} and provisional_name {}".format(band, provisional_name))
        default_comment = ""

        source_cutout = self.model.get_current_cutout()

        if not auto:
            result = self.view.ds9.get('imexam key coordinate wcs fk5 degrees')
            # result = display.get("""imexam key coordinate $x $y $filename""")
            values = result.split()
            ra = Quantity(float(values[1]), unit=units.degree)
            dec = Quantity(float(values[2]), unit=units.degree)
            self.place_marker(ra, dec, radius=2 * units.arcsec, colour='green')
            key = values[0]
        else:
            key = isinstance(auto, bool) and " " or auto
            ra = source_cutout.ra
            dec = source_cutout.dec

        (x, y, extno) = source_cutout.world2pix(ra, dec, usepv=False)
        source_cutout.update_pixel_location((float(x), float(y)), extno)

        pre_daophot_pixel_x = source_cutout.pixel_x
        pre_daophot_pixel_y = source_cutout.pixel_y

        # self.view.mark_apertures(source_cutout, pixel=False)

        try:
            phot = self.model.get_current_source_observed_magnitude()
            cen_x = phot['XCENTER'][0]
            cen_y = phot['YCENTER'][0]
            obs_mag = phot['MAG'][0]
            obs_mag_err = phot['MERR'][0]
            phot_failure = phot['PIER'][0] != 0
            sky_failure = phot['SIER'][0] != 0
            cen_failure = phot['CIER'][0] != 0
        except Exception as er:
            logger.critical("PHOT ERROR: {}".format(er))
            phot_failure = sky_failure = cen_failure = True
            obs_mag = ""
            cen_x = pre_daophot_pixel_x
            cen_y = pre_daophot_pixel_y
            obs_mag_err = -1
            band = ""
            default_comment = str(er)

        obs_mag = phot_failure and None or obs_mag
        obs_mag_err = phot_failure and None or obs_mag_err
        source_cutout.update_pixel_location((cen_x, cen_y), extno=extno)
        source_cutout._adjusted = False

        if math.sqrt((cen_x - pre_daophot_pixel_x) ** 2 + (cen_y - pre_daophot_pixel_y) ** 2) > 1.5 or cen_failure:
            # check if the user wants to use the 'hand' coordinates or these new ones.
            self.place_marker(cen_x, cen_y, 6, colour='white')
            self.view.show_offset_source_dialog((pre_daophot_pixel_x, pre_daophot_pixel_y), (cen_x, cen_y))

        note1_default = ""
        if self.model.is_current_source_adjusted():
            note1_default = config.read("MPC.NOTE1_HAND_ADJUSTED")
        else:
            for note in config.read("MPC.NOTE1OPTIONS"):
                if note.lower().startswith(key):
                    note1_default = note
                    break
        note1 = len(note1_default) > 0 and note1_default[0] or note1_default

        if isinstance(self, ProcessTracksController):
            this_observation = mpc.Observation(
                null_observation=False,
                provisional_name=provisional_name,
                note1=note1,
                note2=config.read('MPC.NOTE2DEFAULT')[0],
                date=self.model.get_current_observation_date(),
                ra=self.model.get_current_ra(),
                dec=self.model.get_current_dec(),
                mag=obs_mag,
                mag_err=obs_mag_err,
                band=band,
                observatory_code=config.read("MPC.DEFAULT_OBSERVATORY_CODE"),
                discovery=self.is_discovery,
                comment=None,
                xpos=source_cutout.observed_x,
                ypos=source_cutout.observed_y,
                frame=self.model.get_current_reading().obs.rawname,
                astrometric_level=source_cutout.astrom_header.get('ASTLEVEL', None))

            try:
                previous_observations = self.model.get_writer().get_chronological_buffered_observations()
                for idx, observation in enumerate(previous_observations):
                    try:
                        if observation.comment.frame.strip() == this_observation.comment.frame.strip():
                            previous_observations[idx] = this_observation
                            this_observation = False
                            break
                    except Exception as ex:
                        print type(ex), str(ex)
                if this_observation:
                    previous_observations.append(this_observation)
                print Orbfit(previous_observations).summarize()
            except Exception as ex:
                logger.error(type(ex), str(ex))
                print "Failed to compute preliminary orbit."

        if obs_mag < 24 and auto is not False:
            self.on_do_accept(None,
                              provisional_name,
                              sky_failure and "S  poor sky" or note1_default,
                              config.read("MPC.NOTE2DEFAULT"),
                              self.model.get_current_observation_date(),
                              self.model.get_current_ra(),
                              self.model.get_current_dec(),
                              obs_mag,
                              obs_mag_err,
                              band,
                              config.read("MPC.DEFAULT_OBSERVATORY_CODE"), ""
                              )
        else:
            self.view.show_accept_source_dialog(
                provisional_name,
                self.model.get_current_observation_date(),
                self.model.get_current_ra(),
                self.model.get_current_dec(),
                obs_mag,
                obs_mag_err,
                band,
                note1_choices=config.read("MPC.NOTE1OPTIONS"),
                note2_choices=config.read("MPC.NOTE2OPTIONS"),
                note1_default=sky_failure and "S  poor sky" or note1_default,
                note2_default=config.read("MPC.NOTE2DEFAULT"),
                default_observatory_code=config.read("MPC.DEFAULT_OBSERVATORY_CODE"),
                default_comment=default_comment,
                phot_failure=phot_failure,
                pixel_x=source_cutout.pixel_x,
                pixel_y=source_cutout.pixel_y)

    def on_do_accept(self,
                     minor_planet_number,
                     provisional_name,
                     note1,
                     note2,
                     date_of_obs,
                     ra,
                     dec,
                     obs_mag,
                     obs_mag_err,
                     band,
                     observatory_code,
                     comment):
        """
        Final acceptance with collected data.
        """
        # Just extract the character code from the notes, not the
        # full description
        note1_code = note1.split(" ")[0]
        note2_code = note2.split(" ")[0]

        self.view.close_accept_source_dialog()

        self.model.set_current_source_name(provisional_name)

        reading = self.model.get_current_reading()
        source_cutout = self.model.get_current_cutout()

        mpc_observation = mpc.Observation(
            null_observation=False,
            provisional_name=provisional_name,
            note1=note1_code,
            note2=note2_code,
            date=date_of_obs,
            ra=ra,
            dec=dec,
            mag=obs_mag,
            mag_err=obs_mag_err,
            band=band,
            observatory_code=observatory_code,
            discovery=self.is_discovery,
            comment=comment,
            xpos=source_cutout.observed_x,
            ypos=source_cutout.observed_y,
            frame=reading.obs.rawname,
            astrometric_level=source_cutout.astrom_header.get('ASTLEVEL', None)
        )
        mpc_observation._date_precision = 6

        source_reading = self.model.get_current_reading()
        source_reading.sky_coord = mpc_observation.coordinate
        source_reading.pix_coord = mpc_observation.comment.x, mpc_observation.comment.y

        data = self.model.get_current_workunit().data
        key = mpc_observation.comment.frame.strip()
        data.mpc_observations[key] = mpc_observation

        self.model.get_writer().write(mpc_observation)

        self.model.accept_current_item()
        reset_frame = False
        if self.model.get_current_workunit().get_current_source_readings().is_on_last_item():
            self.view.clear()
            reset_frame = True
        self.model.next_item()
        if reset_frame:
            self.view.frame(1)

    def on_cancel_accept(self):
        self.view.close_accept_source_dialog()

    def on_reject(self):
        self.view.show_reject_source_dialog()

    def on_do_offset(self, cen_coords):
        self.view.close_offset_source_dialog()

        source_cutout = self.model.get_current_cutout()
        assert isinstance(source_cutout, SourceCutout)
        source_cutout.update_pixel_location(cen_coords, source_cutout.original_observed_ext)
        source_cutout._adjusted = False

    def on_cancel_offset(self, pix_coords):
        self.view.close_offset_source_dialog()

        source_cutout = self.model.get_current_cutout()
        source_cutout.update_pixel_location(pix_coords)

    def on_do_reject(self, comment):
        self.view.close_reject_source_dialog()

        if not self.model.is_current_source_named():
            self.model.set_current_source_name(self._generate_provisional_name())

        reading = self.model.get_current_reading()

        mpc_observation = mpc.Observation(
            provisional_name=self.model.get_current_source_name(),
            date=self.model.get_current_observation_date(),
            ra=self.model.get_current_ra(),
            dec=self.model.get_current_dec(),
            xpos=reading.x,
            ypos=reading.y,
            frame=reading.obs.rawname,
            comment=comment)
        mpc_observation._date_precision = 6

        mpc_observation.null_observation = True

        self.model.get_writer().write(mpc_observation)

        self.model.reject_current_item()
        self.model.next_item()

    def on_cancel_reject(self):
        self.view.close_reject_source_dialog()


class ProcessCandidatesController(AbstractController):
    def on_accept(self):
        writer = self.model.get_writer()
        writer.write_source(self.model.get_current_source())

        self.model.accept_current_item()
        self.view.clear()
        self.model.next_item()

    def on_reject(self):
        self.model.reject_current_item()
        self.view.clear()
        self.model.next_item()

    def on_load_comparison(self, research=False):
        pass


class ProcessTracksController(ProcessRealsController):
    """
    The main controller of the 'track' task.  Sets up the view and
    handles user interactions. This task extends orbit linkages from
    three out to more observations.
    """

    def __init__(self, model, view, name_generator):
        super(ProcessTracksController, self).__init__(model, view, name_generator)
        assert isinstance(model, TransAckValidationModel)
        assert isinstance(view, ApplicationView)
        self.use_pixel_coords = False
        self.mark_prediction = True
        self.is_discovery = False
        self.minimum_align_offset = 60


    def on_accept(self, auto=False):
        super(ProcessTracksController, self).on_accept(auto=False)

    def on_do_accept(self,
                     minor_planet_number,
                     provisional_name,
                     note1,
                     note2,
                     date_of_obs,
                     ra,
                     dec,
                     obs_mag,
                     obs_mag_err,
                     band,
                     observatory_code,
                     comment):
        super(ProcessTracksController, self).on_do_accept(
            minor_planet_number,
            provisional_name,
            note1,
            note2,
            date_of_obs,
            ra,
            dec,
            obs_mag,
            obs_mag_err,
            band,
            observatory_code,
            comment)

        try:
            self.model.get_current_workunit().print_orbfit_info()
        except OrbfitError as error:
            logger.error("Orbfit Error: {0}".format(error))
        self.is_discovery = False

    def on_load_comparison(self, research=False):
        """
        Display the comparison image
        """

        logger.debug(str(research))
        cutout = self.model.get_current_cutout()
        if research or cutout.comparison_image is None:
            cutout.retrieve_comparison_image()
        if cutout.comparison_image is not None:  # if a comparison image was found
            self.view.display(cutout.comparison_image, self.use_pixel_coords)
            self.view.align(self.model.get_current_cutout(),
                            self.model.get_current_reading(),
                            self.model.get_current_source())
            self.model.get_current_workunit().previous_obs()
            self.model.acknowledge_image_displayed()

    def on_ssos(self):
        try:
            new_workunit = self.model.get_current_workunit().query_ssos()
            self.model.add_workunit(new_workunit)
        except AssertionError as e:
            logger.critical(str(e))
            pass
        self.model.next_item()

    def on_save(self):
        workunit = self.model.get_current_workunit()
        assert isinstance(workunit, TracksWorkUnit)
        print "Saved to: {}".format(workunit.save())


class ImageLoadingDialogManager(object):
    def __init__(self, view):
        self.view = view
        self._wait_items = set()
        self._dialog_showing = False

    def wait_for_item(self, item):
        self._wait_items.add(item)

        if not self._dialog_showing:
            self.view.show_image_loading_dialog()
            self._dialog_showing = True

    def set_item_done(self, item):
        if item not in self._wait_items:
            return

        self._wait_items.remove(item)

        if len(self._wait_items) == 0 and self._dialog_showing:
            self.view.hide_image_loading_dialog()
            self._dialog_showing = False



__author__ = "David Rusk <drusk@uvic.ca>"
import Queue
import threading
from ..gui import logger
from ..gui import config

MAX_THREADS = config.read('APP.MAX_THREADS')


class AsynchronousDownloadManager(object):
    """
    Coordinates the downloading of images asynchronously from the rest of
    the application.
    """

    def __init__(self, downloader, error_handler):
        """
        Constructor.

        Args:
          downloader:
            Downloads images.
          error_handler:
            Handles errors that occur when trying to download resources.
        """
        self.downloader = downloader
        self.error_handler = error_handler

        self._work_queue = Queue.Queue()

        self._workers = []
        self._maximize_workers()

    def submit_request(self, request):
        self._work_queue.put(request)
        self._maximize_workers()

    def stop_download(self):
        for worker in self._workers:
            worker.stop()

    def wait_for_downloads_to_stop(self):
        while not self._all_workers_stopped():
            pass

    def refresh_vos_client(self):
        self.downloader.refresh_vos_client()

    def _maximize_workers(self):
        self._prune_dead_workers()

        while len(self._workers) < MAX_THREADS:
            worker = DownloadThread(self._work_queue, self.downloader,
                                    self.error_handler)
            worker.daemon = True  # Thread quits when application does
            self._workers.append(worker)
            worker.start()

            logger.debug("Created download worker thread (total=%d)" %
                         len(self._workers))

    def _prune_dead_workers(self):
        self._workers = filter(lambda thread: thread.is_alive(), self._workers)

    def _all_workers_stopped(self):
        for worker in self._workers:
            if not worker.is_stopped():
                return False

        return True


class DownloadRequest(object):
    """
    Specifies an item (image and potentially related files) to be downloaded.
    """

    def __init__(self,
                 reading,
                 focus=None,
                 needs_apcor=False,
                 callback=None):
        """
        Constructor.

        Args:
          source_reading: ossos.astrom.SourceReading
            The reading which will be the focus of the downloaded image.
          focus: tuple(int, int)
            The x, y coordinates that should be the focus of the downloaded
            image.  These coordinates should be in terms of the
            source_reading parameter's coordinate system.
            Default value is None, in which case the source reading's x, y
            position is used as the focus.
          needs_apcor: bool
            If True, the apcor file with data needed for photometry
            calculations is downloaded in addition to the image.
            Defaults to False.
          callback: callable
            An optional callback to be called with the downloaded snapshot
            as its argument.
        """
        self.reading = reading
        self.needs_apcor = needs_apcor
        self.callback = callback

        if focus is None:
            self.focus = reading.source_point
        else:
            self.focus = focus

    def execute(self, downloader):
        cutout = downloader.download_cutout(self.reading,
                                            focus=self.focus,
                                            needs_apcor=self.needs_apcor)
        logger.debug("Got cutout: {}".format(cutout))
        if self.callback is not None:
            self.callback(cutout)


class DownloadThread(threading.Thread):
    def __init__(self, work_queue, downloader, error_handler):
        super(DownloadThread, self).__init__()

        self.work_queue = work_queue
        self.downloader = downloader
        self.error_handler = error_handler

        self._should_stop = False
        self._idle = True

    def run(self):
        while not self._should_stop:
            download_request = self.work_queue.get()
            self._idle = False

            try:
                self.do_download(download_request)
            except Exception as error:
                self.error_handler.handle_error(error, download_request)
            finally:
                # It is up to the error handler to requeue the downloadable
                # item if needed.
                self.work_queue.task_done()
                self._idle = True

    def do_download(self, download_request):
        download_request.execute(self.downloader)

    def stop(self):
        self._should_stop = True

    def is_stopped(self):
        return self._should_stop and self._idle



__author__ = "David Rusk <drusk@uvic.ca>"


class CutoutGrid(object):
    """
    A grid of cutouts with frames being rows and times being columns.
    """

    def __init__(self, source):
        self.source = source
        self.num_frames = source.num_readings()
        self.num_times = source.num_readings()

        self._grid = [[None] * self.num_frames for _ in range(self.num_times)]

    @property
    def shape(self):
        return self.num_frames, self.num_times

    def apply(self, function):
        for frame_index in range(self.num_frames):
            for time_index in range(self.num_times):
                function(self.get_cutout(frame_index, time_index),
                         frame_index, time_index)

    def add_cutout(self, cutout, frame_index, time_index):
        self._grid[frame_index][time_index] = cutout

    def get_cutout(self, frame_index, time_index):
        return self._grid[frame_index][time_index]

    def get_hdulist(self, frame_index, time_index):
        return self.get_cutout(frame_index, time_index).hdulist

    def get_hdulists(self, frame_index):
        return map(lambda cutout: cutout.hdulist, self._grid[frame_index])

    def is_filled(self):
        return all(element is not None
                   for frame in self._grid for element in frame)

    def reset_source_location(self):
        def reset_cutout(cutout, frame_index, time_index):
            cutout.reset_source_location()

        self.apply(reset_cutout)



"""OSSOS VOSpace storage convenience package"""
import cStringIO
import errno
import fnmatch
from glob import glob
import os
import re
from string import upper
import tempfile
import logging
import warnings

from astropy.coordinates import SkyCoord
from astropy.io import ascii
from astropy import units
from astropy.units import Quantity
from astropy.utils.exceptions import AstropyUserWarning
from vos import vos
from astropy.io import fits
import requests

requests.packages.urllib3.disable_warnings()
requests_log = logging.getLogger("requests.packages.urllib3")
requests_log.setLevel(logging.ERROR)


from .downloads.cutouts.calculator import CoordinateConverter

import coding
import util
from .gui import logger
from .wcs import WCS

MAXCOUNT = 30000
_TARGET = "TARGET"

CERTFILE = os.path.join(os.getenv('HOME'),
                        '.ssl',
                        'cadcproxy.pem')

DBIMAGES = 'vos:OSSOS/dbimages'
MEASURE3 = 'vos:OSSOS/measure3'
POSTAGE_STAMPS = 'vos:OSSOS/postage_stamps'
TRIPLETS = 'vos:OSSOS/triplets'
ASTROM_RELEASES = 'vos:OSSOS/0_OSSOSreleases'

DATA_WEB_SERVICE = 'https://www.canfar.phys.uvic.ca/data/pub/'
VOSPACE_WEB_SERVICE = 'https://www.canfar.phys.uvic.ca/vospace/nodes/'
TAP_WEB_SERVICE = 'http://www.cadc-ccda.hia-iha.nrc-cnrc.gc.ca/tap/sync'

OSSOS_TAG_URI_BASE = 'ivo://canfar.uvic.ca/ossos'
OBJECT_COUNT = "object_count"


class Wrapper(object):

    def __init__(self, wrapped_class, *args, **kargs):
        self.wrapped_class = wrapped_class(*args, **kargs)

    def __getattr__(self, attr):
        orig_attr = self.wrapped_class.__getattribute__(attr)
        if callable(orig_attr):
            def hooked(*args, **kwargs):
                print "-->", orig_attr, args, kwargs
                result = orig_attr(*args, **kwargs)
                print "<--", type(result)
                return result
            return hooked
        else:
            return orig_attr


# vospace = Wrapper(vos.Client)
vospace = vos.Client()

SUCCESS = 'success'

# ## some cache holders.
mopheaders = {}
astheaders = {}
fwhm = {}
zmag = {}

APCOR_EXT = "apcor"
ZEROPOINT_USED_EXT = "zeropoint.used"
PSF_EXT = "psf.fits"


def get_detections(release='current'):
    """

    @param release: the release to retrieve from VOSpace
    @return: astropy.table.Table
    """
    detection_path = "{}/{}/OSSOS*.detections".format(ASTROM_RELEASES,release)
    filenames = vospace.glob(detection_path)
    if not len(filenames) > 0:
        raise IOError("No detection file found using: {}".format(detection_path))

    return ascii.read(open_vos_or_local(filenames[0]).read(), header_start=-1)


def cone_search(ra, dec, dra=0.01, ddec=0.01, mjdate=None, calibration_level=2):
    """Do a QUERY on the TAP service for all observations that are part of OSSOS (*P05/*P016)
    where taken after mjd and have calibration 'observable'.

    :param ra: RA center of search cont
    :type ra: Quantity
    :param dec: float degrees
    :type dec: Quantity
    :param dra: float degrees
    :type dra: Quantity
    :param ddec: float degrees
    :type ddec: Quantity
    """

    data = dict(QUERY=(" SELECT Observation.observationID as collectionID, "
                       " Plane.time_bounds_cval1 AS mjdate "
                       " FROM caom2.Observation AS Observation "
                       " JOIN caom2.Plane AS Plane "
                       " ON Observation.obsID = Plane.obsID "
                       " WHERE  ( Observation.collection = 'CFHT' ) "
                       " AND Plane.calibrationLevel={} "
                       " AND Plane.energy_bandpassName LIKE 'r.%' "
                       " AND ( Observation.proposal_id LIKE '%P05' or Observation.proposal_id LIKE '%P06' )"
                       " AND Observation.target_name NOT LIKE 'WP%'"),
                REQUEST="doQuery",
                LANG="ADQL",
                FORMAT="tsv")

    data["QUERY"] = data["QUERY"].format(calibration_level)
    data["QUERY"] += (" AND  "
                      " CONTAINS( BOX('ICRS', {}, {}, {}, {}), "
                      " Plane.position_bounds ) = 1 ").format(ra.to(units.degree).value, dec.to(units.degree).value,
                                                              dra.to(units.degree).value, ddec.to(units.degree).value)
    if mjdate is not None:
        data["QUERY"] += " AND Plane.time_bounds_cval1 < {} AND Plane.time_bounds_cval2 > {} ".format(
            mjdate + 1.0 / 24.0,
            mjdate - 1 / 24.0)

    result = requests.get(TAP_WEB_SERVICE, params=data, verify=False)
    assert isinstance(result, requests.Response)
    logger.debug("Doing TAP Query using url: %s" % (str(result.url)))

    table_reader = ascii.get_reader(Reader=ascii.Basic)
    table_reader.header.splitter.delimiter = '\t'
    table_reader.data.splitter.delimiter = '\t'
    table = table_reader.read(result.text)

    logger.debug(str(table))
    return table


def populate(dataset_name,
             data_web_service_url=DATA_WEB_SERVICE + "CFHT"):
    """Given a dataset_name created the desired dbimages directories
    and links to the raw data files stored at CADC."""

    data_dest = get_uri(dataset_name, version='o', ext='fits.fz')
    data_source = "%s/%so.fits.fz" % (data_web_service_url, dataset_name)

    mkdir(os.path.dirname(data_dest))

    try:
        vospace.link(data_source, data_dest)
    except IOError as e:
        if e.errno == errno.EEXIST:
            pass
        else:
            raise e

    header_dest = get_uri(dataset_name, version='o', ext='head')
    header_source = "%s/%so.fits.fz?cutout=[0]" % (
        data_web_service_url, dataset_name)
    try:
        vospace.link(header_source, header_dest)
    except IOError as e:
        if e.errno == errno.EEXIST:
            pass
        else:
            raise e

    header_dest = get_uri(dataset_name, version='p', ext='head')
    header_source = "%s/%s/%sp.head" % (
        'http://www.cadc-ccda.hia-iha.nrc-cnrc.gc.ca/data/pub', 'CFHTSG', dataset_name)
    try:
        vospace.link(header_source, header_dest)
    except IOError as e:
        if e.errno == errno.EEXIST:
            pass
        else:
            raise e

    return True


def get_cands_uri(field, ccd, version='p', ext='measure3.cands.astrom', prefix=None,
                  block=None):
    """
    return the nominal URI for a candidate file.
    """

    if prefix is None:
        prefix = ""
    if len(prefix) > 0:
        prefix += "_"
    if len(field) > 0:
        field += "_"

    if ext is None:
        ext = ""
    if len(ext) > 0 and ext[0] != ".":
        ext = ".{}".format(ext)

    measure3_dir = MEASURE3
    if block is not None:
        measure3_dir + "/{}".format(block)
    mkdir(measure3_dir)
    return "{}/{}{}{}{}{}".format(measure3_dir, prefix, field, version, ccd, ext)


def get_uri(expnum, ccd=None,
            version='p', ext='fits',
            subdir=None, prefix=None):
    """
    Build the uri for an OSSOS image stored in the dbimages
    containerNode.

    :rtype : basestring
    expnum: CFHT exposure number
    ccd: CCD in the mosaic [0-35]
    version: one of p,s,o etc.
    dbimages: dbimages containerNode.
    @type subdir: str
    @param expnum: int
    @param ccd:
    @param version:
    @param ext:
    @param subdir:
    @param prefix:
    """
    if subdir is None:
        subdir = str(expnum)
    if prefix is None:
        prefix = ''
    uri = os.path.join(DBIMAGES, subdir)

    if ext is None:
        ext = ''
    elif len(ext) > 0 and ext[0] != '.':
        ext = '.' + ext

    if version is None:
        version = ''

    if ccd is None:
        uri = os.path.join(uri,
                           '%s%s%s%s' % (prefix, str(expnum),
                                         version,
                                         ext))
    else:
        ccd = str(ccd).zfill(2)
        uri = os.path.join(uri,
                           'ccd{}'.format(ccd),
                           '%s%s%s%s%s' % (prefix, str(expnum),
                                           version,
                                           ccd,
                                           ext))

    return uri


dbimages_uri = get_uri


def set_tags_on_uri(uri, keys, values=None):
    node = vospace.get_node(uri)
    if values is None:
        values = []
        for idx in range(len(keys)):
            values.append(None)
    assert (len(values) == len(keys))
    for idx in range(len(keys)):
        key = keys[idx]
        value = values[idx]
        tag = tag_uri(key)
        node.props[tag] = value
    return vospace.add_props(node)


def _set_tags(expnum, keys, values=None):
    uri = os.path.join(DBIMAGES, str(expnum))
    node = vospace.get_node(uri, force=True)
    if values is None:
        values = []
        for idx in range(len(keys)):
            values.append(None)
    assert (len(values) == len(keys))
    for idx in range(len(keys)):
        key = keys[idx]
        tag = tag_uri(key)
        value = values[idx]
        node.props[tag] = value
    return vospace.add_props(node)


def set_tags(expnum, props):
    """Assign the key/value pairs in props as tags on on the given expnum.

    @param expnum: str
    @param props: dict
    @return: success
    """
    # now set all the props
    return _set_tags(expnum, props.keys(), props.values())


def set_tag(expnum, key, value):
    """Assign a key/value pair tag to the given expnum containerNode.

    @param expnum:  str
    @param key: str
    @param value: str
    @return: success
    """

    return set_tags(expnum, {key: value})


def tag_uri(key):
    """Build the uri for a given tag key. 

    key: (str) eg 'mkpsf_00'

    """
    if OSSOS_TAG_URI_BASE in key:
        return key
    return OSSOS_TAG_URI_BASE + "#" + key.strip()


def get_tag(expnum, key):
    """given a key, return the vospace tag value.

    @return: value of the tag requested.
    """

    if tag_uri(key) not in get_tags(expnum):
        get_tags(expnum, force=True)
    logger.debug("%s # %s -> %s" % (expnum, tag_uri(key), get_tags(expnum).get(tag_uri(key), None)))
    return get_tags(expnum).get(tag_uri(key), None)


def get_process_tag(program, ccd, version='p'):
    """make a process tag have a suffix indicating which ccd its for.

    """
    return "%s_%s%s" % (program, str(version), str(ccd).zfill(2))


def get_tags(expnum, force=False):
    """

    @param expnum:
    @param force:
    @return: a dictionary of tags.
    @rtype: dict
    """
    uri = os.path.join(DBIMAGES, str(expnum))
    return vospace.get_node(uri, force=force).props


class Task(object):
    """
    A task within the OSSOS pipeline work-flow.
    """

    def __init__(self, executable, dependency=None):
        self.executable = executable
        self.name = os.path.splitext(self.executable)[0]
        self._target = None
        self._status = None
        self._dependency = None
        self.dependency = dependency

    def __str__(self):
        return self.name

    @property
    def tag(self):
        """
        Get the string representation of the tag used to annotate the status in VOSpace.
        @return: str
        """
        return "{}{}_{}{}{:02d}".format(self.target.prefix,
                                        self,
                                        self.target.version,
                                        self.target.ccd)

    @property
    def target(self):
        """

        @return: The target that this task is set to run on.
        @rtype: Target
        """
        return self._target

    @target.setter
    def target(self, target):
        assert isinstance(Target, target)
        self._target = target

    @property
    def dependency(self):
        """
        @rtype: Task
        """
        raise NotImplementedError()

    @dependency.setter
    def dependency(self, dependency):
        if dependency is None:
            self._dependency = dependency
        else:
            assert isinstance(Task, dependency)

    @property
    def status(self):
        """

        @return: The status of running this task on the given target.
        @rtype: str
        """
        return get_tag(self.target.expnum, self.tag)

    @status.setter
    def status(self, status):
        status += util.Time.now().iso
        set_tag(self.target.expnum, self.tag, status)

    @property
    def finished(self):
        """
        @rtype: bool
        """
        return self.status.startswith(SUCCESS)

    @property
    def ready(self):
        if self.dependency is None:
            return True
        else:
            return self.dependency.finished


class Target(object):
    """
    The target that a task will act on.
    """

    def __init__(self, prefix, expnum, version, ccd):
        """

        @param prefix: that is prefixed to the base exposure
        @type prefix: str
        @param expnum: the number of the CFHT exposure that is the target
        @type expnum: str
        @param version: Which version of the exposure (o, p, s) is the target.
        @type version: str
        @param ccd: which CCD of the exposure is the target.
        @type ccd: int
        """

        self.prefix = prefix
        self.expnum = expnum
        self.version = version
        self.ccd = ccd

    @property
    def name(self):
        return "{}{}{}{:02d}".format(self.prefix, self.expnum, self.version, self.ccd)

    @property
    def tags(self):
        """

        @rtype: list [str]
        """
        return get_tags(self.expnum)


def get_status(task, prefix, expnum, version, ccd, return_message=False):
    """Report back status of the given program.

    @param prefix:
    """
    key = get_process_tag(prefix+task, ccd, version)
    status = get_tag(expnum, key)
    logger.debug('%s: %s' % (key, status))
    if return_message:
        return status
    else:
        return status == SUCCESS


def set_status(task, prefix, expnum, version, ccd, status):
    """set the processing status of the given program.

    @param prefix:
    """

    return set_tag(expnum, get_process_tag(prefix+task, ccd, version), status)


def get_file(expnum, ccd=None, version='p', ext='fits', subdir=None, prefix=None):
    uri = get_uri(expnum=expnum, ccd=ccd, version=version, ext=ext, subdir=subdir, prefix=prefix)
    filename = os.path.basename(uri)

    if not os.access(filename, os.F_OK):
        copy(uri, filename)

    return filename


def decompose_content_decomposition(content_decomposition):
    """

    :param content_decomposition:
    :return:
    """
    # check for '-*' in the cutout string and replace is naxis:1

    content_decomposition = re.findall('(\d+)__(\d+)_(\d+)_(\d+)_(\d+)', content_decomposition)
    if len(content_decomposition) == 0:
        content_decomposition = [(0, 1, -1, 1, -1)]
    return content_decomposition


def ra_dec_cutout(uri, sky_coord, radius):
    """

    :param uri: The vospace location of the image to make a cutout from
    :type uri: str
    :param sky_coord: The central coordinate of the cutout
    :type sky_coord: SkyCoord
    :param radius: The radius of the cutout
    :type radius: Quantity
    :return: An HDUList with the cutout
    :rtype: fits.HDUList
    """

    # These two lines enable debugging at httplib level (requests->urllib3->http.client)
    # You will see the REQUEST, including HEADERS and DATA, and RESPONSE with HEADERS but without DATA.
    # The only thing missing will be the response.body which is not logged.
    try:
        import http.client as http_client
    except ImportError:
        # Python 2
        import httplib as http_client

    # Get the 'uncut' images CRPIX1/CRPIX2 values

    coo_sys = upper(sky_coord.frame.name)
    this_cutout = "CIRCLE {} {} {} {}".format(coo_sys, sky_coord.ra.to(units.degree).value,
                                              sky_coord.dec.to(units.degree).value,
                                              radius.to(units.degree).value)

    view = "cutout"

    fobj = vospace.open(uri, view=view, cutout=this_cutout)
    hdulist = fits.open(cStringIO.StringIO(fobj.read()))
    hdulist.verify('silentfix+ignore')
    cutouts = decompose_content_decomposition(fobj.resp.headers.get('content-disposition', '0___'))
    logger.debug("Got cutout boundaries of: {}".format(cutouts))
    logger.debug("Initial Length of HDUList: {}".format(len(hdulist)))

    # Make sure here is a primaryHDU
    if len(hdulist) == 1:
        phdu = fits.PrimaryHDU()
        phdu.header['ORIGIN'] = "OSSOS"
        hdulist.insert(0, phdu)

    logger.debug("Final Length of HDUList: {}".format(len(hdulist)))

    if len(cutouts) != len(hdulist) - 1:
        raise ValueError("Wrong number of cutout structures found in Content-Disposition response.")

    for hdu in hdulist[1:]:
        cutout = cutouts.pop(0)
        if 'ASTLEVEL' not in hdu.header:
            print "******* NO ASTLEVEL ********* ASTROMETRY NOT POSSIBLE ********* IGNORE {0} ********".format(uri)
        hdu.header['EXTNO'] = cutout[0]
        hdu.header['DATASEC'] = reset_datasec("[{}:{},{}:{}]".format(cutout[1],
                                                                     cutout[2],
                                                                     cutout[3],
                                                                     cutout[4]),
                                              hdu.header.get('DATASEC', None),
                                              hdu.header['NAXIS1'],
                                              hdu.header['NAXIS2'])
        hdu.header['XOFFSET'] = int(cutout[1]) - 1
        hdu.header['YOFFSET'] = int(cutout[3]) - 1

        hdu.converter = CoordinateConverter(hdu.header['XOFFSET'], hdu.header['YOFFSET'])
        try:
                hdu.wcs = WCS(hdu.header)
        except Exception as ex:
            logger.error("Failed trying to initialize the WCS for {}".format(uri))
            raise ex
    logger.debug("Sending back {}".format(hdulist))
    return hdulist


def get_image(expnum, ccd=None, version='p', ext='fits',
              subdir=None, prefix=None, cutout=None, return_file=True, flip_image=True):
    """Get a FITS file for this expnum/ccd  from VOSpace.


    @param cutout: (str)
    @param return_file: return an filename (True) or HDUList (False)
    @param expnum: CFHT exposure number (int)
    @param ccd:     @param ccd:
    @param version: [p, s, o]  (char)
    @param ext:
    @param subdir:
    @param prefix:
    @return: astropy.io.fits.PrimaryHDU
    """

    filename = os.path.basename(get_uri(expnum, ccd, version, ext=ext, subdir=subdir, prefix=prefix))
    if os.access(filename, os.F_OK) and return_file and cutout is None:
        return filename

    cutout_string = cutout
    try:
        if os.access(filename, os.F_OK) and cutout:
            cutout = datasec_to_list(cutout)
            hdulist = fits.open(filename)
            if len(hdulist) > 1:
                raise ValueError("Local cutout access not designed to work on MEFs yet.")
            header = hdulist[0].header
            cutout[0] = cutout[0] < 0 and header['NAXIS1'] - cutout[0] + 1 or cutout[0]
            cutout[1] = cutout[1] < 0 and header['NAXIS1'] - cutout[1] + 1 or cutout[1]
            cutout[2] = cutout[2] < 0 and header['NAXIS2'] - cutout[2] + 1 or cutout[2]
            cutout[3] = cutout[3] < 0 and header['NAXIS2'] - cutout[3] + 1 or cutout[3]
            logger.debug("DATA array shape: {}".format(hdulist[0].data.shape))
            logger.debug("CUTOUT array: {} {} {} {}".format(cutout[0],
                                                            cutout[1],
                                                            cutout[2],
                                                            cutout[3]))
            flip = cutout[0] < cutout[1] and 1 or -1
            flop = cutout[2] < cutout[3] and 1 or -1
            header['CRPIX1'] = (header.get("CRPIX1", 1) - cutout[0]) * flip + 1
            header['CRPIX2'] = (header.get("CRPIX2", 1) - cutout[2]) * flop + 1
            header['CD1_1'] = header.get("CD1_1", 1) * flip
            header['CD2_1'] = header.get("CD2_1", 1) * flip
            header['CD2_2'] = header.get("CD2_2", 1) * flop
            header['CD1_2'] = header.get("CD1_2", 1) * flop

            data = hdulist[0].data[cutout[2]-1:cutout[3], cutout[0]-1:cutout[1]]
            hdulist[0].data = data
            header['DATASEC'] = reset_datasec(cutout_string,
                                              header['DATASEC'],
                                              header['NAXIS1'],
                                              header['NAXIS2'])
            if return_file:
                cutout_filename = os.path.splitext(filename)[0]+"_{}_{}_{}_{}.fits".format(cutout[0],
                                                                                           cutout[1],
                                                                                           cutout[2],
                                                                                           cutout[3])
                hdulist.writeto(cutout_filename)
                return cutout_filename
            else:
                return hdulist
    except Exception as e:
        logger.debug(str(e))
        logger.debug("Failed trying to access local copy: {} with cutout [{}:{}, {}:{}], using VOSpace".format(
            filename, cutout[2]-1, cutout[3], cutout[0]-1, cutout[1]))

    cutout = cutout_string

    if not subdir:
        subdir = str(expnum)

    logger.debug("Building list of possible uri locations")
    # # here is the list of places we will look, in order
    if version != 'p':
        locations = [(get_uri(expnum, ccd, version, ext=ext, subdir=subdir, prefix=prefix),
                      cutout)]
    else:
        locations = []
    logger.debug(str(locations))
    if ccd is not None:
        try:
            for this_ext in [ext, ext + ".fz"]:
                ext_no = int(ccd) + 1
                # extension 1 -> 18 +  37,36 should be flipped.
                flip_these_extensions = range(1,19)
                flip_these_extensions.append(37)
                flip_these_extensions.append(38)
                flip = (cutout is None and "fits" in ext and (
                    (ext_no in flip_these_extensions and flip_image) and "[-*,-*]" or "[*,*]")) or cutout
                locations.append((get_uri(expnum, version=version, ext=this_ext, subdir=subdir),
                                  "[{}]{}".format(ext_no, flip)))
        except Exception as e:
            logger.error(str(e))
            pass
    else:
        uri = get_uri(expnum, ccd, version, ext=ext, subdir=subdir, prefix=prefix)
        locations.append((uri, cutout))
        uri = get_uri(expnum, ccd, version, ext=ext + ".fz", subdir=subdir, prefix=prefix)
        locations.append((uri, cutout))
    while len(locations) > 0:
        (uri, cutout) = locations.pop(0)
        try:
            hdu_list = get_hdu(uri, cutout)
            if return_file:
                hdu_list.writeto(filename)
                del hdu_list
                return filename
            else:
                return hdu_list
        except Exception as e:
            logger.debug("{}".format(type(e)))
            logger.debug("Failed to open {} cutout:{}".format(uri, cutout))
            logger.debug("vos sent back error: {} code: {}".format(str(e), getattr(e, 'errno', 0)))

    raise IOError(errno.ENOENT, "Failed to get image using {} {} {} {}.".format(expnum, version, ccd, cutout))


def datasec_to_list(datasec):
    """
    convert an IRAF style PIXEL DATA section as to a list of integers.
    @param datasec: str
    @return: list
    """

    return [int(x) for x in re.findall(r"([-+]?[\*\d]+?)[:,\]]+", datasec)]


def reset_datasec(cutout, datasec, naxis1, naxis2):
    """
    reset the datasec to account for a possible cutout.

    @param cutout:
    @param datasec:
    @return:
    """

    if cutout is None or cutout == "[*,*]":
        return datasec

    try:
        datasec = datasec_to_list(datasec)
    except:
        return datasec

    # check for '-*' in the cutout string and replace is naxis:1
    cutout = cutout.replace(" ", "")
    cutout = cutout.replace("[-*,", "{}:1,".format(naxis1))
    cutout = cutout.replace(",-*]", ",{}:1]".format(naxis2))
    cutout = cutout.replace("[*,", "[1:{},".format(naxis1))
    cutout = cutout.replace(",*]", ",1:{}]".format(naxis1))

    try:
        cutout = [int(x) for x in re.findall(r"([-+]?[\*\d]+?)[:,\]]+", cutout)]
    except:
        logger.debug("Failed to processes the cutout pattern: {}".format(cutout))
        return datasec

    if len(cutout) == 5:
        # cutout likely starts with extension, remove
        cutout = cutout[1:]

    # -ve integer offsets indicate offset from the end of array.
    for idx in [0, 1]:
        if cutout[idx] < 0:
            cutout[idx] = naxis1 - cutout[idx] + 1
    for idx in [2, 3]:
        if cutout[idx] < 0:
            cutout[idx] = naxis2 - cutout[idx] + 1

    flip = cutout[0] > cutout[1]
    flop = cutout[2] > cutout[3]

    logger.debug("Working with cutout: {}".format(cutout))

    if flip:
        cutout = [naxis1 - cutout[0] + 1, naxis1 - cutout[1] + 1, cutout[2], cutout[3]]
        datasec = [naxis1 - datasec[1] + 1, naxis1 - datasec[0] + 1, datasec[2], datasec[3]]

    if flop:
        cutout = [cutout[0], cutout[1], naxis2 - cutout[2] + 1, naxis2 - cutout[3] + 1]
        datasec = [datasec[0], datasec[1], naxis2 - datasec[3] + 1, naxis2 - datasec[2] + 1]

    datasec = [max(datasec[0] - cutout[0] + 1, 1),
               min(datasec[1] - cutout[0] + 1, naxis1),
               max(datasec[2] - cutout[2] + 1, 1),
               min(datasec[3] - cutout[2] + 1, naxis2)]

    return "[{}:{},{}:{}]".format(datasec[0], datasec[1], datasec[2], datasec[3])


def get_hdu(uri, cutout=None):
    """Get a at the given uri from VOSpace, possibly doing a cutout.

    If the cutout is flips the image then we also must flip the datasec keywords.  Also, we must offset the
    datasec to reflect the cutout area being used.

    @return: fits.HDU
    """

    # the filename is based on the Simple FITS images file.
    filename = os.path.basename(uri)
    if os.access(filename, os.F_OK) and cutout is None:
        logger.debug("File already on disk: {}".format(filename))
        hdu_list = fits.open(filename, scale_back=True)
        hdu_list.verify('silentfix+ignore')

    else:
        logger.debug("Pulling: {}{} from VOSpace".format(uri, cutout))
        fpt = tempfile.NamedTemporaryFile(suffix='.fits')
        cutout = cutout is not None and cutout or ""
        vospace.copy(uri+cutout, fpt.name)
        fpt.seek(0, 2)
        fpt.seek(0)
        logger.debug("Read from vospace completed. Building fits object.")
        hdu_list = fits.open(fpt, scale_back=False)
        hdu_list.verify('silentfix+ignore')

        logger.debug("Got image from vospace")
        try:
            hdu_list[0].header['DATASEC'] = reset_datasec(cutout, hdu_list[0].header['DATASEC'],
                                                          hdu_list[0].header['NAXIS1'],
                                                          hdu_list[0].header['NAXIS2'])
        except Exception as e:
            logging.debug("error converting datasec: {}".format(str(e)))

    for hdu in hdu_list:
        logging.debug("Adding converter to {}".format(hdu))
        hdu.converter = CoordinateConverter(0, 0)
        try:
            hdu.wcs = WCS(hdu.header)
        except Exception as ex:
            logger.error("Failed trying to initialize the WCS: {}".format(ex))

    return hdu_list


def get_trans(expnum, ccd, prefix=None, version='p'):
    uri = get_uri(expnum, ccd, version, ext='trans.jmp', prefix=prefix)
    logging.info("get_trans: {}".format(uri))
    fobj = open_vos_or_local(uri)
    line = fobj.read()
    fobj.close()
    vs = line.split()
    trans = {'dx': float(vs[0]),
             'cd11': float(vs[1]),
             'cd12': float(vs[2]),
             'dy': float(vs[3]),
             'cd21': float(vs[4]),
             'cd22': float(vs[5])}
    return trans


def get_fwhm(expnum, ccd, prefix=None, version='p'):
    """Get the FWHM computed for the given expnum/ccd combo.

    @param expnum:
    @param ccd:
    @param prefix:
    @param version:
    @return:
    """

    uri = get_uri(expnum, ccd, version, ext='fwhm', prefix=prefix)
    filename = os.path.basename(uri)

    try:
        return fwhm[uri]
    except:
        pass

    try:
        fwhm[uri]=float(open(filename, 'r').read())
        return fwhm[uri]
    except:
        pass

    try:
        fwhm[uri] = float(open_vos_or_local(uri).read())
        return fwhm[uri]
    except Exception as e:
        fwhm[uri] = 4.0
        return fwhm[uri]



def get_zeropoint(expnum, ccd, prefix=None, version='p'):
    """Get the zeropoint for this exposure.

    This command expects that there is a file called #######p##.zeropoint.used which contains the zeropoint.

    @param expnum: exposure to get zeropoint of
    @param ccd: which ccd (extension - 1) to get zp
    @param prefix: possible string prefixed to expsoure number.
    @param version:  one of [spo] as in #######p##
    @return:
    """
    uri = get_uri(expnum, ccd, version, ext='zeropoint.used', prefix=prefix)
    try:
        return zmag[uri]
    except:
        pass

    try:
        zmag[uri] = float(open_vos_or_local(uri).read())
        return zmag[uri]
    except:
        pass

    zmag[uri] = 0.0
    return zmag[uri]


def mkdir(dirname):
    """make directory tree in vospace.

    @param dirname: name of the directory to make
    """
    dir_list = []

    while not vospace.isdir(dirname):
        dir_list.append(dirname)
        dirname = os.path.dirname(dirname)
    while len(dir_list) > 0:
        logging.info("Creating directory: %s" % (dir_list[-1]))
        try:
            vospace.mkdir(dir_list.pop())
        except IOError as e:
            if e.errno == errno.EEXIST:
                pass
            else:
                raise e


def vofile(filename, **kwargs):
    """Open and return a handle on a VOSpace data connection"""
    basename = os.path.basename(filename)
    if os.access(basename, os.R_OK):
        return open(basename, 'r')
    kwargs['view'] = kwargs.get('view', 'data')
    return vospace.open(filename, **kwargs)


def open_vos_or_local(path, mode="rb"):
    """
    Opens a file which can either be in VOSpace or the local filesystem.
    """
    if path.startswith("vos:"):
        primary_mode = mode[0]
        if primary_mode == "r":
            vofile_mode = os.O_RDONLY
        elif primary_mode == "w":
            vofile_mode = os.O_WRONLY
        elif primary_mode == "a":
            vofile_mode = os.O_APPEND
        else:
            raise ValueError("Can't open with mode %s" % mode)

        return vofile(path, mode=vofile_mode)
    else:
        return open(path, mode)


def copy(source, dest):
    """use the vospace service to get a file. """

    #logger.debug("deleting {} ".format(dest))
    #try:
    #    vospace.delete(dest)
    #except Exception as e:
    #    logger.debug(str(e))
    #    pass
    logger.info("copying {} -> {}".format(source, dest))

    return vospace.copy(source, dest)


def vlink(s_expnum, s_ccd, s_version, s_ext,
          l_expnum, l_ccd, l_version, l_ext, s_prefix=None, l_prefix=None):
    """make a link between two version of a file.

    @param s_expnum:
    @param s_ccd:
    @param s_version:
    @param s_ext:
    @param l_expnum:
    @param l_ccd:
    @param l_version:
    @param l_ext:
    @param s_prefix:
    @param l_prefix:
    @return:
    """
    source_uri = get_uri(s_expnum, ccd=s_ccd, version=s_version, ext=s_ext, prefix=s_prefix)
    link_uri = get_uri(l_expnum, ccd=l_ccd, version=l_version, ext=l_ext, prefix=l_prefix)

    return vospace.link(source_uri, link_uri)


def remove(uri):
    try:
        vospace.delete(uri)
    except Exception as e:
        logger.debug(str(e))


def delete(expnum, ccd, version, ext, prefix=None):
    """delete a file, no error on does not exist."""
    uri = get_uri(expnum, ccd=ccd, version=version, ext=ext, prefix=prefix)
    remove(uri)


def my_glob(pattern):
    """get a listing matching pattern"""
    result = []
    if pattern[0:4] == 'vos:':
        dirname = os.path.dirname(pattern)
        flist = listdir(dirname)
        for fname in flist:
            fname = '/'.join([dirname, fname])
            if fnmatch.fnmatch(fname, pattern):
                result.append(fname)
    else:
        result = glob(pattern)
    return result


def listdir(directory, force=False):
    return vospace.listdir(directory, force=force)


def list_dbimages():
    return listdir(DBIMAGES)


def exists(uri, force=False):
    try:
        return vospace.get_node(uri, force=force) is not None
    except EnvironmentError as e:
        logger.error(str(e))  # not critical enough to raise
        # Sometimes the error code returned is the OS version, sometimes the HTTP version
        if e.errno in [404, os.errno.ENOENT]:
            return False


def move(old_uri, new_uri):
    vospace.move(old_uri, new_uri)


def delete_uri(uri):
    vospace.delete(uri)


def has_property(node_uri, property_name, ossos_base=True):
    """
    Checks if a node in VOSpace has the specified property.
    """
    if get_property(node_uri, property_name, ossos_base) is None:
        return False
    else:
        return True


def get_property(node_uri, property_name, ossos_base=True):
    """
    Retrieves the value associated with a property on a node in VOSpace.
    """
    # Must use force or we could have a cached copy of the node from before
    # properties of interest were set/updated.
    node = vospace.get_node(node_uri, force=True)
    property_uri = tag_uri(property_name) if ossos_base else property_name

    if property_uri not in node.props:
        return None

    return node.props[property_uri]


def set_property(node_uri, property_name, property_value, ossos_base=True):
    """
    Sets the value of a property on a node in VOSpace.  If the property
    already has a value then it is first cleared and then set.
    """
    node = vospace.get_node(node_uri)
    property_uri = tag_uri(property_name) if ossos_base else property_name

    # If there is an existing value, clear it first
    if property_uri in node.props:
        node.props[property_uri] = None
        vospace.add_props(node)

    node.props[property_uri] = property_value
    vospace.add_props(node)


def build_counter_tag(epoch_field, dry_run=False):
    """
    Builds the tag for the counter of a given epoch/field,
    without the OSSOS base.
    """
    logger.info("Epoch Field: {}, OBJECT_COUNT {}".format(str(epoch_field), str(OBJECT_COUNT)))
    tag = epoch_field[1] + "-" + OBJECT_COUNT

    if dry_run:
        tag += "-DRYRUN"

    return tag


def read_object_counter(node_uri, epoch_field, dry_run=False):
    """
    Reads the object counter for the given epoch/field on the specified
    node.

    Returns:
      count: str
        The current object count.
    """
    return get_property(node_uri, build_counter_tag(epoch_field, dry_run),
                        ossos_base=True)


def increment_object_counter(node_uri, epoch_field, dry_run=False):
    """
    Increments the object counter for the given epoch/field on the specified
    node.

    Returns:
      new_count: str
        The object count AFTER incrementing.
    """
    current_count = read_object_counter(node_uri, epoch_field, dry_run=dry_run)

    if current_count is None:
        new_count = "01"
    else:
        new_count = coding.base36encode(coding.base36decode(current_count) + 1,
                                        pad_length=2)

    set_property(node_uri,
                 build_counter_tag(epoch_field, dry_run=dry_run),
                 new_count,
                 ossos_base=True)

    return new_count


def get_mopheader(expnum, ccd, version='p', prefix=None):
    """
    Retrieve the mopheader, either from cache or from vospace
    :rtype : fits.Header
    """
    prefix = prefix is None and "" or prefix
    mopheader_uri = dbimages_uri(expnum=expnum,
                                 ccd=ccd,
                                 version=version,
                                 prefix=prefix,
                                 ext='.mopheader')
    if mopheader_uri in mopheaders:
        return mopheaders[mopheader_uri]

    filename = os.path.basename(mopheader_uri)

    if os.access(filename, os.F_OK):
        logger.debug("File already on disk: {}".format(filename))
        mopheader_fpt = cStringIO.StringIO(open(filename, 'r').read())
    else:
        mopheader_fpt = cStringIO.StringIO(open_vos_or_local(mopheader_uri).read())

    with warnings.catch_warnings():
        warnings.simplefilter('ignore', AstropyUserWarning)
        mopheader = fits.open(mopheader_fpt)

        # add some values to the mopheader so it can be an astrom header too.
        header = mopheader[0].header
        try:
            header['FWHM'] = get_fwhm(expnum, ccd)
        except IOError:
            header['FWHM'] = 10
        header['SCALE'] = mopheader[0].header['PIXSCALE']
        header['NAX1'] = header['NAXIS1']
        header['NAX2'] = header['NAXIS2']
        header['MOPversion'] = header['MOP_VER']
        header['MJD_OBS_CENTER'] = str(util.Time(header['MJD-OBSC'],
                                                 format='mjd',
                                                 scale='utc', precision=5).replicate(format='mpc'))
        header['MAXCOUNT'] = MAXCOUNT
        mopheaders[mopheader_uri] = header
        mopheader.close()
    return mopheaders[mopheader_uri]


def _get_sghead(expnum, version):
    """
    Use the data web service to retrieve the stephen's astrometric header.

    :param expnum: CFHT exposure number you want the header for
    :param version: Which version of the header? ('p', 's', 'o')
    :rtype : list of astropy.io.fits.Header objects.
    """

    logger.warning("_get_sghead is depricated, use get_astheader instead.")

    url = "http://www.cadc-ccda.hia-iha.nrc-cnrc.gc.ca/data/pub/CFHTSG/{}{}.head".format(expnum, version)
    logging.getLogger("requests").setLevel(logging.ERROR)
    resp = requests.get(url)
    if resp.status_code != 200:
        raise IOError(errno.ENOENT, "Could not get {}".format(url))

    header_str_list = re.split('END      \n', resp.content)

    # # make the first entry in the list a Null
    headers = [None]
    for header_str in header_str_list:
        headers.append(fits.Header.fromstring(header_str, sep='\n'))

    return headers


def get_header(uri):
    """
    Pull a FITS header from observation at the given URI
    """
    if uri not in astheaders:
        astheaders[uri] = get_hdu(uri, cutout="[1:1,1:1]")[0].header
    return astheaders[uri]


def get_astheader(expnum, ccd, version='p', prefix=None, ext=None):
    """
    Retrieve the header for a given dbimages file.

    @param expnum:  CFHT odometer number
    @param ccd: which ccd based extension (0..35)
    @param version: 'o','p', or 's'
    @return:
    """
    logger.debug("Getting ast header for {}".format(expnum))
    ast_uri = dbimages_uri(expnum, ccd, version=version, ext='.fits')
    if ast_uri not in astheaders:
        hdulist = get_image(expnum, ccd=ccd, version=version, prefix=prefix,
                            cutout="[1:1,1:1]", return_file=False)
        assert isinstance(hdulist, fits.HDUList)
        astheaders[ast_uri] = hdulist[0].header
    return astheaders[ast_uri]


def log_filename(prefix, task, version, ccd):
    return "{}{}_{}{}.txt".format(prefix, task, version, ccd)


def log_location(expnum, ccd):
    return os.path.dirname(get_uri(expnum, ccd=ccd))


def set_logger(task, prefix, expnum, ccd, version, dry_run):
    this_logger = logging.getLogger()
    log_format = logging.Formatter('%(asctime)s - %(module)s.%(funcName)s %(lineno)d: %(message)s')

    filename = log_filename(prefix, task, ccd=ccd, version=version)
    location = log_location(expnum, ccd)
    if not dry_run:
        vo_handler = util.VOFileHandler("/".join([location, filename]))
        vo_handler.setFormatter(log_format)
        this_logger.addHandler(vo_handler)

    file_handler = logging.FileHandler(filename=filename)
    file_handler.setFormatter(log_format)
    this_logger.addHandler(file_handler)



__author__ = "David Rusk <drusk@uvic.ca>"

# NOTE: we are using version 1 of the pubsub API
# (http://pubsub.sourceforge.net/apidocs/docs4v1.html#label-api-v1-docs)
# This is in order to maintain compatiblity with software available on older
# systems such as ScientificLinux 5.5 on CANFAR VMs.

try:
    # On up-to-date systems this import will trigger using version 1 of
    # the pubsub API.
    from wx.lib.pubsub import setupv1
except ImportError:
    # Old systems will not be able to execute that import, but it is ok
    # because all they have is the version 1 API.
    pass

try:
    from wx.lib.pubsub import Publisher as pub
except ImportError:
    from wx.lib.pubsub import setuparg1
    from wx.lib.pubsub import pub

# Event ids
ROOT = ("rootid", )
CHANGE_IMAGE = ROOT + ("change_image", )
IMG_LOADED = ROOT + ("imgload", )
FINISHED_WORKUNIT = ROOT + ("finished_workunit", )
NO_AVAILABLE_WORK = ROOT + ("no_available_work", )


def send(event_id, data=None):
    pub.sendMessage(event_id, data=data)


def subscribe(event_id, callback):
    pub.subscribe(callback, event_id)


def unsub_all():
    pub.unsubAll()



__author__ = 'Michele Bannister   git:@mtbannister'

'''
Plot a single OSSOS object's path on the sky and observations over the course of the survey.
'''

import argparse

from matplotlib.patches import Rectangle
from matplotlib import pyplot as plt
from astropy.time import Time, TimeDelta
# import aplpy

from ossos import mpc
from ossos import orbfit
import sky_location_plots
import parameters
# import parsers
from planning.ObsStatus import query_for_observations


def plot_actual_observations(ax, obj, mark_rejected=False):
    for i, obs in enumerate(obj.mpc_observations):
        fc = 'b'
        alpha = 0.6
        marker = 'o'
        if mark_rejected:
            if obs.null_observation:
                fc = 'r'
                marker = 'x'
        ax.scatter(obs.coordinate.ra.degree, obs.coordinate.dec.degree, marker=marker, facecolor=fc, alpha=alpha)
        # is it away from the last annotation plotted? 5 days empirically worked best to remove puzzlement
        if ((obs.date - obj.mpc_observations[i - 1].date) > TimeDelta(5, format='jd')) or i == 0:
            ax.annotate(obs.date.datetime.strftime('%Y-%m-%d'),
                        (obs.coordinate.ra.degree + .5, obs.coordinate.dec.degree - 0.01),
                        size=7,
                        color='k')
    return ax


def plot_prediction(ax, obj):
    ra = []
    dec = []
    start = Time(parameters.SURVEY_START)  # min([d.date for d in obj.mpc_observations]) - TimeDelta(90, format='jd')
    end = Time('2016-01-01')
    date = start
    orbit = orbfit.Orbfit(obj.mpc_observations)
    while date < end:
        orbit.predict(date)
        ra.append(orbit.coordinate.ra.degree)
        dec.append(orbit.coordinate.dec.degree)
        date = date + TimeDelta(10, format='jd')

    ax.plot(ra, dec, 'r-', linewidth=0.6)

    # some labels for different times
    orbit.predict(start)
    ax.annotate(parameters.SURVEY_START,
                (orbit.coordinate.ra.degree - 0.1, orbit.coordinate.dec.degree),
                size=7,
                color='k')
    orbit.predict(end)
    ax.annotate(end.datetime.strftime('%Y-%m-%d'),
                (orbit.coordinate.ra.degree - 0.1, orbit.coordinate.dec.degree),
                size=7,
                color='k')

    return ax


def megacam_corresponding_to_observations(ax):
    # now while we know the KBO was in these images (here's the measurements), want the field centres
    # to make the footprints.
    discov = [m for m in obj.mpc_observations if m.discovery][0]
    obs_frames = [m.comment.frame.split('p')[0].strip(' ') for m in obj.mpc_observations]
    obs_table = query_for_observations(Time(parameters.SURVEY_START).mjd, 1, tuple(parameters.OSSOS_RUNIDS))

    camera_dimen = 0.98

    for i in range(len(obs_table) - 1, -1, -1):
        t = obs_table.data[i]
        if str(t.dataset_name) in obs_frames:
            print str(t.dataset_name)
            alpha = 0.4
            lw = 0.95
            if Time(t.StartDate, format='mjd') < Time('2014-01-01'):
                ec = 'b'
            else:
                ec = '#E47833'
                alpha = 0.5
            if str(t.dataset_name) == discov.comment.frame[0:7]:
                ec = 'k'
                lw = 1.7
            r = Rectangle(xy=(t.RA - camera_dimen / 2.0, t.DEC - camera_dimen / 2.0),
                          height=camera_dimen,
                          width=camera_dimen,
                          edgecolor=ec,
                          alpha=alpha,
                          lw=lw,
                          zorder=-100,
                          fill=False)
            ax.add_artist(r)

    return ax


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("kbo",
                        nargs='*',
                        help="Plot all OSSOS imaging of the given KBO, which can be an OSSOS discovery.")

    args = parser.parse_args()

    # load in the KBO. For now let's assume it's a single OSSOS discovery, generalise later
    # discoveries = parsers.ossos_release_parser(table=True)
    dir = '/Users/bannisterm/Dropbox/OSSOS/measure3/2014B-H/track/'
    fn = dir + args.kbo[0] + '.ast'  # parameters.REAL_KBO_AST_DIR
    obj = mpc.MPCReader(fn)  # let MPCReader's logic determine the provisional name
    # need to determine a plot extent to get the MegaCam coverage
    # and KBOs go westward so earliest to latest is kinda useful
    extents = [max([obs.coordinate.ra.degree for obs in obj.mpc_observations]) + 0.9,
               min([obs.coordinate.ra.degree for obs in obj.mpc_observations]) - 0.9,
               min([obs.coordinate.dec.degree for obs in obj.mpc_observations]),
               max([obs.coordinate.dec.degree for obs in obj.mpc_observations])]
    midpoint = (extents[0] + ((extents[1] - extents[0]) / 2.), extents[2] + ((extents[3] - extents[2]) / 2.))
    print(extents, midpoint)

    handles, labels, ax, fontP = sky_location_plots.basic_skysurvey_plot_setup()
    plt.axis('equal')
    plt.grid(alpha=0.2)
    #
    # fig = aplpy.FITSFigure()
    # fig.set_theme('publication')

    ax = plot_actual_observations(ax, obj, mark_rejected=True)
    ax = plot_prediction(ax, obj)
    ax = megacam_corresponding_to_observations(ax)

    # only for o3e13 poster child for first-quarter paper
    if args.kbo[0] == 'o3e13':
        ax.annotate('invariable plane', (213.7, -11.9), alpha=0.7, rotation=19)

    sky_location_plots.saving_closeout(args.kbo[0], '', extents, '')



__author__ = 'Michele Bannister   git:@mtbannister'

import os

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
from astropy.table import Table

import prettyplotlib as ppl
import plot_fanciness
from parsers import ossos_discoveries
from parameters import RELEASE_VERSION
from ossos import mpc, orbfit

path = '/Users/bannisterm/Dropbox/Papers in progress/OSSOS/First_quarter/data/'


def cfeps_residuals(version='L7', regenerate=False):
    if version == 'L4':
        header = ['obsnum', 'arclen', 'date', 'ra_resid', 'date2', 'dec_resid']
        data = Table.read(path + 'L4resid.txt',
                          format='ascii', guess=False, delimiter=' ', data_start=0, names=header)
        cfeps_mas = (np.sqrt(data['ra_resid'] ** 2 + data['dec_resid'] ** 2)) * 1000

    else:
        all_resids = []
        if regenerate:
            data_path = '/Users/bannisterm/Dropbox/mpc/'
            for filename in os.listdir(data_path):
                observations = mpc.MPCReader(data_path + filename)
                print observations.provisional_name
                orbit = orbfit.Orbfit(observations.mpc_observations)
                try:
                    residuals = orbit.residuals(overall=True)
                    all_resids += residuals
                except Exception, e:
                    print observations.provisional_name
                    print e
                    continue
            print all_resids

            with open('cfeps_resids_{}.dat'.format(version), 'w') as outfile:
                for r in all_resids:
                    outfile.write("{}\n".format(r))

        else:
            assert os.path.exists('cfeps_resids_{}.dat'.format(version))
            with open('cfeps_resids_{}.dat'.format(version), 'r') as infile:
                lines = infile.readlines()
                for line in lines:
                    all_resids.append(float(line))
            print len(all_resids)

    cfeps_mas = np.array(all_resids)
    print cfeps_mas
    print 'cfeps', np.median(cfeps_mas)

    return cfeps_mas


def ossos_residuals(version=RELEASE_VERSION, regenerate=False):
    all_resids = []  # arcsec

    if regenerate:
        # takes a few minutes to compute all these
        discoveries = ossos_discoveries(data_release=version)  # gets the full MPC lines for the current release
        for obj in discoveries:
            print obj.name
            residuals = obj.orbit.residuals(overall=True)
            all_resids += residuals
        print all_resids

        with open('ossos_resids_v{}.dat'.format(RELEASE_VERSION), 'w') as outfile:
            for r in all_resids:
                outfile.write("{}\n".format(r))

    else:
        assert os.path.exists('ossos_resids_v{}.dat'.format(RELEASE_VERSION))
        with open('ossos_resids_v{}.dat'.format(RELEASE_VERSION), 'r') as infile:
            lines = infile.readlines()
            for line in lines:
                all_resids.append(float(line))
        print len(all_resids)

    ossos_mas = np.array(all_resids)
    print ossos_mas
    print 'ossos', np.median(ossos_mas)

    return ossos_mas


if __name__ == '__main__':
    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(5, 5))

    cfeps_mas = cfeps_residuals(version='L7')
    ossos_mas = ossos_residuals()

    bins = 16
    maxrange = .85
    ax.hist(cfeps_mas,
            histtype='step', bins=bins, range=(0, maxrange),  # these seem appropriate
            color='r', linestyle='solid', lw=2.1, alpha=0.3, label='CFEPS L7')
    ax.hist(ossos_mas,
            histtype='step', bins=bins, range=(0, maxrange),
            color='b', linestyle='solid', label='OSSOS 13A')

    ax.grid(True, alpha=0.3)
    plt.xlabel("astrometric residual (arcsec)")
    plt.ylabel("number of astrometric measurements")
    plt.xlim([0, .85])

    handler_map = [
        mlines.Line2D([], [], color='r', linestyle='solid', lw=2.1, alpha=0.3, label='CFEPS L7'),
        mlines.Line2D([], [], color='b', label='OSSOS 13A'),
    ]

    ppl.legend(title="Survey",
               fontsize='small', handletextpad=0.5)
    plot_fanciness.remove_border(ax)

    plt.draw()
    outfile = path.rpartition('data')[0] + 'figures/astrometric_residual_50mas_bins.pdf'
    plt.savefig(outfile, transparent=True, bbox_inches='tight')



from ossos.gui import logger

__author__ = "David Rusk <drusk@uvic.ca>"


class CoordinateConverter(object):
    def __init__(self, x_offset, y_offset, invert=False):
        """

        @type y_offset: float
        @type x_offset: float
        """
        self._dx = x_offset
        self._dy = y_offset
        self._invert = invert and -1 or 1
        logger.debug("Convert initialized as x_offset, y_offset {},{}".format(self.x_offset, self.y_offset))

    @property
    def x_offset(self):
        """
        Amount to subtract from x position to move to new reference frame.

        @rtype: float
        """
        return self._invert * self._dx

    @property
    def y_offset(self):
        """
        Amount to subtract from the y position to move to the new reference frame.

        @rtype: float
        """
        return self._invert * self._dy

    def convert(self, point):
        """
        Convert a point from one coordinate system to another.

        Args:
          point: tuple(int x, int y)
            The point in the original coordinate system.

        Returns:
          converted_point: tuple(int x, int y)
            The point in the new coordinate system.

        Example: convert coordinate from original image into a pixel location
          within a cutout image.

        @rtype: list(float,float)
        """
        x, y = point
        (x1, y1) = x - self.x_offset, y - self.y_offset
        logger.debug("converted {} {} ==> {} {}".format(x, y, x1, y1))
        return x1, y1

    def get_inverse_converter(self):
        """
        Returns a converter object for converting back from this converter's
        output coordinate system to its input coordinate system.

        @rtype : CoordinateConverter
        """

        return CoordinateConverter(self.x_offset, self.y_offset, invert=True)


__author__ = "David Rusk <drusk@uvic.ca>"

CANDS_TASK = "cands"
REALS_TASK = "reals"
TRACK_TASK = "track"

task_list = [CANDS_TASK, REALS_TASK, TRACK_TASK]

suffixes = {
    CANDS_TASK: ".cands.astrom",
    REALS_TASK: ".reals.astrom",
    TRACK_TASK: ".mpc"
}


def get_suffix(task):
    return suffixes[task]



__author__ = 'Michele Bannister   git:@mtbannister'

import math
import os
import datetime

import numpy as np
import ephem
import matplotlib.pyplot as plt

import parameters
import mpcread
import parsers
import horizons







# FIXME: make this an args setting rather than hardwired
PLOT_MPCORB = True and os.access(parameters.MPCORB_FILE, os.F_OK)

saturn_moons = ['Phoebe', 'Ymir', 'Paaliaq', 'Tarvos', 'Ijiraq', 'Suttungr', 'Kiviuq', 'Mundilfari',
                'Albiorix', 'Skathi', 'Erriapus', 'Siarnaq', 'Thrymr', 'Narvi', 'Bestla', 'Hyrrokkin', 'Kari']
# pulled out 'Iapetus', on Saturn


def plot_planets(ax, plot, date, hill_sphere=False):
    mass = {"Sun": 1.989 * 10 ** 30, "Mars": 639 * 10 ** 21, "Jupiter": 1.898 * 10 ** 27, "Saturn": 568.3 * 10 ** 24,
            "Uranus": 86.81 * 10 ** 24, "Neptune": 102.4 * 10 ** 24}  # kg
    for planet in [ephem.Mars(), ephem.Jupiter(), ephem.Saturn(), ephem.Uranus(), ephem.Neptune()]:
        planet.compute(ephem.date(date))
        pos = (math.degrees(planet.ra), math.degrees(planet.dec))
        # if plot_polygon.isInside(math.degrees(planet.ra), math.degrees(planet.dec)):
        ax.scatter(pos[0], pos[1],
                   marker='o',
                   s=30,
                   facecolor='k',  # #E47833',
                   edgecolor='k')  # #E47833')
        ax.annotate(planet.name, (pos[0] - .4, pos[1] + 0.1))  # (pos[0]+.9, pos[1]+0.5))  # offset to make it readable

        if hill_sphere:
            print planet.name, planet.sun_distance, mass[planet.name],
            hs_radius = (planet.sun_distance * ephem.meters_per_au) * (
                (mass[planet.name] / 3 * mass['Sun']) ** (1 / 3.))
            angular_size = planet.earth_distance * hs_radius  # FIXME
            print 'Hill sphere', hs_radius, hs_radius / ephem.meters_per_au, angular_size
            ax.add_patch(plt.Circle(pos, radius=angular_size, fill=False))

    return ax


def plot_saturn_moons(ax):
    for s_moon in saturn_moons:  # get Saturn's irregular moons
        # 'Kiviuq' to left. Hyr mag 24.3, Bestla mag 24.5, Kiviuq 22.6, Iji 23.3
        close = ['Ijiraq', 'Bestla', 'Hyrrokkin', 'Kiviuq']
        ax = plot_single_known_tno(ax, s_moon, "2013-04-09 08:50", close=close)

    return ax


def plot_ossos_discoveries(ax, discoveries, blockname, prediction_date=False):
    fc = ['b', '#E47833']
    alpha = [0.6, 1]
    marker = ['o', 'd']
    size = [25, 30]
    pl_index = np.where((discoveries['cl'] == 'res') & (discoveries['j'] == 3) & (discoveries['k'] == 2))
    l = []
    for k, n in enumerate(discoveries):
        if k not in pl_index[0]:
            l.append(k)
    not_plutinos = discoveries[l]
    plutinos = discoveries[pl_index]
    for j, d in enumerate([not_plutinos, plutinos]):
        print len(d)
        ra = [math.degrees(ephem.degrees(ephem.hours(str(n)))) for n in d['ra_dis']]
        dec = [float(q) for q in d['dec_dis']]
        ax.scatter(ra, dec,
                   marker=marker[j], s=size[j], facecolor=fc[j], edgecolor='k', linewidth=0.4,
                   alpha=alpha[j])  # original size=4.5
        for kbo in d:
            ra = math.degrees(ephem.degrees(ephem.hours(str(kbo['ra_dis']))))
            dec = kbo['dec_dis']
            name = kbo['object'].rpartition('o3' + blockname[-1].lower())[2]
            if j == 1:
                sep = 0.20
            else:
                sep = 0.18
            if name.endswith('PD'):
                latsep = .2
            else:
                latsep = .07
            ax.annotate(name,
                        (ra - latsep, dec - sep),
                        size=9,
                        color='k')

    return ax

    # for kbo in discoveries:
    #     # If the provided date isn't very close to the object's intrinsic discovery date,
    #     # predict the object's position for the provided date.
    #     if prediction_date:  # In this case, have to use prediction as data release doesn't have Orbfit.orbit property
    #         # Gets around field discovery times differing by a small fraction of an hour
    #         pd = time.Time(prediction_date.replace('/', '-'))
    #         if abs(kbo.discovery.date - pd) > time.TimeDelta(0.5, format='jd'):
    #             kbo.orbit.predict(prediction_date.replace('/', '-'))
    #             ra = kbo.orbit.coordinate.ra.degree
    #             dec = kbo.orbit.coordinate.dec.degree
    #         else:  # close enough we can just use the discovery's time
    #             ra = kbo.discovery.coordinate.ra.degree
    #             dec = kbo.discovery.coordinate.dec.degree
    #     else:  # no complexity: use discovery locations from data-release
    #         assert (kbo.ra_discov is not None) and (kbo.dec_discov is not None)
    #         ra = math.degrees(ephem.degrees(ephem.hours(str(kbo.ra_discov))))
    #         dec = math.degrees(ephem.degrees(str(kbo.dec_discov)))
    #
    #     assert ra, dec

        # assert kbo.classification is not None
        # if (kbo.classification == 'res' and kbo.n == 3 and kbo.m == 2):
        # print kbo.name
        #     fc = '#E47833'
        #     alpha = 1
        # else:
    # fc = 'b'
    # alpha = 0.6
    #
    # ax.scatter(ra, dec, marker='o', facecolor=fc, alpha=alpha, edgecolor='k', linewidth=0.4, s=25)



def plot_known_tnos_singly(ax, extent, date):
    rate_cut = 'a > 15'
    print "PLOTTING LOCATIONS OF KNOWN KBOs (using {})".format(parameters.MPCORB_FILE)
    kbos = mpcread.getKBOs(parameters.MPCORB_FILE)
    print('Known KBOs: {}'.format(len(kbos)))
    retkbos = []
    for kbo in kbos:
        kbo.compute(date)
        # # keep only the ones that'd make it onto this plot
        if not ((extent[0] <= math.degrees(kbo.ra) <= extent[1])
                and (extent[2] <= math.degrees(kbo.dec) <= extent[3])):
            continue
        pos = (math.degrees(kbo.ra), math.degrees(kbo.dec))
        ax.scatter(pos[0],
                   pos[1],
                   marker='x',
                   facecolor='r')
        if len(kbo.name) >= 10:  # sorted out for Sofia Pro Light
            ra_shift = -0.45
        else:
            ra_shift = -0.22
        ax.annotate(kbo.name, (pos[0] + ra_shift, pos[1] + 0.06), size=7, color='r')
        retkbos.append(kbo)
    print('Retained KBOs: {}'.format(len(retkbos)))
    return ax, retkbos


def plot_known_tnos_batch(handles, labels, date):
    rate_cut = 'a > 15'
    if os.access(MPCORB, os.F_OK):
        kbos = mpcread.getKBOs(MPCORB, cond=rate_cut)
        kbo_ra = []
        kbo_dec = []
        for kbo in kbos:
            kbo.compute(ephem.date(date))
            kbo_ra.append(math.degrees(kbo.ra))
            kbo_dec.append(math.degrees(kbo.dec))
            ax.scatter(kbo_ra, kbo_dec,
                       marker='+', facecolor='none', edgecolor='g')

    return ax


def plot_single_known_tno(ax, name, date, close=[]):
    # date formatted as %Y-%m-%d %H:%M
    # FIXME: rewrite this to use astroquery.mpc's single-object retrieval.
    dateplus1hr = (datetime.datetime.strptime(date, '%Y-%m-%d %H:%M') + datetime.timedelta(1 / 24.)).strftime(
        '%Y-%m-%d %H:%M')
    obj_elems, single_ephem = horizons.batch(name, date, dateplus1hr, None, su='d')  # pull back one position only
    ra = math.degrees(ephem.degrees(ephem.hours(single_ephem[0]['RA'])))
    dec = math.degrees(ephem.degrees(single_ephem[0]['DEC']))
    if name == 'Ijiraq':
        fc = ec = 'k'
    else:
        fc = ec = 'r'
    ax.scatter(ra,
               dec,
               marker='.',
               facecolor=fc,
               edgecolor=ec,
               alpha=0.4)
    if name in close:  # keep the names from falling on top of each other
        if name == 'Hyrrokkin':
            ax.annotate(name, (ra - .07, dec + .03), size='5', color='r')
        if name == 'Ijiraq':
            ax.annotate(name, (ra - .25, dec + .03), size='5', color='k')
        if name == 'Bestla':
            ax.annotate(name, (ra - .05, dec + .03), size='5', color='r')
        if name == 'Kiviuq':
            ax.annotate(name, (ra, dec + .03), size='5', color='r')
    else:
        ax.annotate(name, (ra - 0.2, dec + .03), size='5', color='r')

    return ax


def plot_synthetic_kbos(ax, coverage):
    ra, dec, hlat, Hmag = parsers.synthetic_model_kbos(kbotype='resonant', arrays=True, maglimit=24.7)
    ax.scatter(ra, dec, c='k', marker='.', s=1, alpha=0.8)

    return ax
